#line 239 "nws/basevec.nw"
#ifndef _NKLEIN_BASEVEC_H_
#define _NKLEIN_BASEVEC_H_ 1

#include <iostream>
#include <sstream>
#include <vector>

    
#line 212 "nws/basevec.nw"
template < typename TT > class basevec;
#line 247 "nws/basevec.nw"
    
#line 125 "nws/basevec.nw"
template < typename TT >
    std::istream& operator >> ( std::istream&, basevec< TT >& );
template < typename TT >
    std::ostream& operator << ( std::ostream&, const basevec< TT >& );

#line 249 "nws/basevec.nw"
    
#line 215 "nws/basevec.nw"
template < typename TT >
class basevec : public std::vector< TT > {
public:
    
#line 24 "nws/basevec.nw"
basevec();
#line 35 "nws/basevec.nw"
basevec( const TT array[], unsigned int len );
#line 50 "nws/basevec.nw"
basevec( const TT value, unsigned int len );
#line 66 "nws/basevec.nw"
basevec( const basevec< TT >& other );
#line 79 "nws/basevec.nw"
const basevec< TT >& operator = ( const basevec< TT >& other );
#line 101 "nws/basevec.nw"
basevec( const char* str );
#line 131 "nws/basevec.nw"
friend std::istream& operator >> <TT> ( std::istream&, basevec< TT >& );
friend std::ostream& operator << <TT> ( std::ostream&, const basevec< TT >& );
#line 219 "nws/basevec.nw"
public:
    static const char* id;
};
#line 250 "nws/basevec.nw"
    
#line 27 "nws/basevec.nw"
template < typename TT >
basevec< TT >::basevec() : std::vector< TT >()
{
}
#line 38 "nws/basevec.nw"
template < typename TT >
basevec< TT >::basevec( const TT array[], unsigned int len )
    : std::vector< TT >( len )
{
    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] = array[ ii ];
    }
}
#line 53 "nws/basevec.nw"
template < typename TT >
basevec< TT >::basevec( const TT value, unsigned int len )
    : std::vector< TT >( len )
{
    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] = value;
    }
}
#line 69 "nws/basevec.nw"
template < typename TT >
basevec< TT >::basevec( const basevec< TT >& other )
    : std::vector< TT >()
{
    *this = other;
}
#line 82 "nws/basevec.nw"
template < typename TT >
const basevec< TT >&
basevec< TT >::operator = ( const basevec< TT >& other )
{
    unsigned int len( other.size() );
    this->resize( len );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] = other[ ii ];
    }

    return *this;
}
#line 104 "nws/basevec.nw"
template < typename TT >
basevec< TT >::basevec( const char* str )
    : std::vector< TT >()
{
    std::istringstream in( str );
    in >> (*this);
    if ( ! in.good() ) {
	this->resize( 0 );
    }
}
#line 142 "nws/basevec.nw"
template < typename TT >
std::istream&
operator >> ( std::istream& in, basevec< TT >& vv )
{
    char bogus;
    unsigned int len;

    in.setf( std::ios::skipws );

    in >> bogus;		// "["
    in >> len;
    in >> bogus >> bogus;	// "]" "<"

    TT zero( 0U );
    vv.resize( len, zero );

    for (unsigned ii=0; ii < len; ++ii) {
	in >> vv[ii];

	if (ii < len-1) {
	    in >> bogus; // ","
	}
    }

    while (bogus != '>' && in.good() ) {
	in >> bogus; // ">"
    }

    return in;
}
#line 180 "nws/basevec.nw"
template < typename TT >
std::ostream&
operator << ( std::ostream& out, const basevec< TT >& vv )
{
    unsigned int len = vv.size();

    out.put( '[' );
    out << len;
    out.put( ']' ).put( '<' ).put( ' ' );

    for (unsigned int ii=0; ii < len; ++ii) {
	out << vv[ii];

	if (ii < len-1) {
	    out.put( ',' ).put( ' ' );
	}
    }

    out.put( ' ' ).put( '>' );

    return out;
}
#line 251 "nws/basevec.nw"
#endif /*_NKLEIN_BASEVEC_H_*/
