#line 442 "nws/color.nw"
#include <string>
#include <sstream>
#include <math.h>
#include "input.h"
#include "color.h"
#include "portal.h"

const char* color::id = "NKVERSION: color( 2.1.2007.03.21 )"; 

#line 127 "nws/color.nw"
color::color( const scene* ss ) : myScene( ss ), 
#line 116 "nws/color.nw"
specularness( 0.1 ), transparency( 0.0 ), reflectiveness( 0.0 ),
phong( 15.0 ), indexOfRefraction( 1.0 ),
usesPosition( false ), usesNormal( false ), usesDirection( false )
#line 128 "nws/color.nw"
{
}
#line 139 "nws/color.nw"
color::color( const char* str, const scene* ss )
    : myScene( ss ), 
#line 116 "nws/color.nw"
specularness( 0.1 ), transparency( 0.0 ), reflectiveness( 0.0 ),
phong( 15.0 ), indexOfRefraction( 1.0 ),
usesPosition( false ), usesNormal( false ), usesDirection( false )
#line 141 "nws/color.nw"
{
    std::istringstream in( str );
    in >> *this;
}
#line 159 "nws/color.nw"
void
color::setDimensions( unsigned int dd )
{
    this->scale.resize( dd, 1.0 );
    ::orthogonalize( this->orientation, dd );
    this->center.resize( dd, 0.0 );

    PortalMap::iterator it;
    for ( it = this->portals.begin(); it != this->portals.end(); ++it ) {
	it->second->setDimensions( dd );
    }
}
#line 216 "nws/color.nw"
void
color::evaluate(
    const mathvec< double >& position,
    const mathvec< double >& normal,
    const mathvec< double >& direction,
    double contribution, unsigned int height,
    mathvec< double >& pixel
) const {
    
#line 234 "nws/color.nw"
if ( pixel.size() < this->channels.size() ) {
    pixel.resize( this->channels.size() );
}
#line 243 "nws/color.nw"
if ( height == 0 || contribution < 0.0001 || this->channels.size() == 0 ) {
    pixel *= 0.0;
    return;
}
#line 225 "nws/color.nw"
    
#line 252 "nws/color.nw"
mathvec<double> pos;
mathvec<double> norm;
mathvec<double> dir;
#line 259 "nws/color.nw"
if ( this->usesPosition ) {
    pos = position * this->scale;
    pos.rotate( this->orientation );
    pos += this->center;
}
#line 270 "nws/color.nw"
if ( this->usesNormal ) {
    norm = normal * this->scale;
    norm.rotate( this->orientation );
    norm.normalize();
}
#line 278 "nws/color.nw"
if ( this->usesDirection ) {
    dir = direction * this->scale;
    dir.rotate( this->orientation );
    dir.normalize();
}
#line 226 "nws/color.nw"
    
#line 286 "nws/color.nw"
    PortalMap& pps( *(PortalMap*)&this->portals );
    PortalMap::iterator it;
    for ( it = pps.begin(); it != pps.end(); ++it ) {
	it->second->evaluate(
	    pos, dir, contribution, height, this->channels.size()
	);
    }
#line 227 "nws/color.nw"
    
#line 300 "nws/color.nw"
double lastVal = 0.0;

for ( unsigned int ii=0; ii < this->channels.size(); ++ii ) {
    pixel[ii] = this->channels[ii]->evaluate( pos, norm, dir ) * contribution;
    lastVal = pixel[ ii ];
}

for ( unsigned int ii=this->channels.size(); ii < pixel.size(); ++ii ) {
    pixel[ii] = lastVal;
}
#line 228 "nws/color.nw"
}
#line 320 "nws/color.nw"
std::istream&
operator >> ( std::istream& stream, color& cc )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "portal" ) {
	    
#line 370 "nws/color.nw"
if ( in.get( token, sizeof(token) ) == 0 ) {
    break;
}
std::string name( token );
portal* pp = new portal( cc.myScene );
stream >> *pp >> semicolon;
cc.portals[ name ] = pp;
#line 337 "nws/color.nw"
	} else if ( tok == "channels" || tok == "color" ) {
	    expr::read( stream, cc.channels, cc.portals );
	    
#line 383 "nws/color.nw"
cc.usesPosition = false;
cc.usesNormal = false;
cc.usesDirection = false;
for ( unsigned int ii=0; ii < cc.channels.size(); ++ii ) {
    cc.usesPosition  = cc.usesPosition  || cc.channels[ ii ]->usesPosition();
    cc.usesNormal    = cc.usesNormal    || cc.channels[ ii ]->usesNormal();
    cc.usesDirection = cc.usesDirection || cc.channels[ ii ]->usesDirection();
}
#line 340 "nws/color.nw"
	    stream >> semicolon;
	} else if ( tok == "specularness" ) {
	    stream >> cc.specularness >> semicolon;
	} else if ( tok == "transparency" ) {
	    stream >> cc.transparency >> semicolon;
	} else if ( tok == "reflectiveness" ) {
	    stream >> cc.reflectiveness >> semicolon;
	} else if ( tok == "phong" ) {
	    stream >> cc.phong >> semicolon;
	} else if ( tok == "indexOfRefraction" || tok == "ior" ) {
	    stream >> cc.indexOfRefraction >> semicolon;
	} else if ( tok == "scale" ) {
	    stream >> cc.scale >> semicolon;
	} else if ( tok == "orientation" ) {
	    stream >> cc.orientation >> semicolon;
	} else if ( tok == "center" ) {
	    stream >> cc.center >> semicolon;
	} else {
	    in.ignore( "COLOR", tok.c_str() );
	}
    }

    return stream;
}
