#line 305 "nws/complement.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "complement.h"

const char* complement::id = "NKVERSION: complement( 2.1.2007.03.21 )"; 

#line 49 "nws/complement.nw"
complement::complement( const scene* ss )
    : object( ss ) 
#line 38 "nws/complement.nw"
, obj( 0 )
#line 51 "nws/complement.nw"
{
}
#line 63 "nws/complement.nw"
complement::~complement( void ) {
    delete this->obj;
};
#line 77 "nws/complement.nw"
void
complement::setDimensions( unsigned int dd )
{
    if ( this->obj != 0 ) {
	this->obj->setDimensions( dd );
    }
    this->object::setDimensions( dd );
}
#line 101 "nws/complement.nw"
bool
complement::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    if ( this->obj == 0 ) {
	return false;
    }
    
#line 120 "nws/complement.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;
#line 111 "nws/complement.nw"
    bool ret = this->obj->intersect( rd, rp, hit );
    if ( ret ) {
	
#line 131 "nws/complement.nw"
hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );
if ( this->gotColor ) {
    hit.recolor( this->chroma );
}
if ( hit.isInside() ) {
    hit.setOutside();
} else {
    hit.setInside();
}
#line 114 "nws/complement.nw"
    }
    return ret;
}
#line 156 "nws/complement.nw"
bool
complement::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    if ( this->obj == 0 ) {
	return false;
    }
    
#line 120 "nws/complement.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;
#line 166 "nws/complement.nw"
    bool ret = this->obj->intersect( rd, rp, hits );
    if ( ret ) {
	for ( unsigned int ii=0; ii < hits.size(); ++ii ) {
	    intersection& hit( *hits[ ii ] );
	    
#line 131 "nws/complement.nw"
hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );
if ( this->gotColor ) {
    hit.recolor( this->chroma );
}
if ( hit.isInside() ) {
    hit.setOutside();
} else {
    hit.setInside();
}
#line 171 "nws/complement.nw"
	}
    } else {
	intersection* nn = new intersection();
	this->forgeHit( direction, start, *nn );
	hits.insert( hits.end(), nn );
    }

    return true;
}
#line 191 "nws/complement.nw"
bool
complement::inside( const mathvec< double >& start ) const
{
    if ( this->obj == 0 ) {
	return false;
    }
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
complement::_inside( const mathvec< double >& rp ) const
{
    if ( this->obj == 0 ) {
	return false;
    }
    return ! this->obj->inside( rp );
}
#line 222 "nws/complement.nw"
std::istream&
operator >> ( std::istream& stream, complement& cc )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "object" ) {
	    object* nn = object::read( stream, cc.myScene );
	    if ( nn != 0 ) {
		delete cc.obj;
		cc.obj = nn;
	    } else {
		std::cerr << "COMPLEMENT got no object!" << std::endl;
	    }
	} else if ( tok == "base" ) {
	    stream >> (object&)cc >> semicolon;
	} else {
	    in.ignore( "COMPLEMENT", tok.c_str() );
	}
    }

    return stream;
}
