#line 601 "nws/convex.nw"
#include <string>
#include <sstream>
#include <list>
#include <math.h>
#include "input.h"
#include "cylinder.h"
#include "halfspace.h"
#include "inter.h"
#include "convex.h"

const char* convex::id = "NKVERSION: convex( 2.1.2007.03.30 )"; 

#line 64 "nws/convex.nw"
convex::convex( const scene* ss ) : csginter( ss ), colorDistance( 0.0 )
{
}
#line 77 "nws/convex.nw"
convex::~convex( void ) {
};
#line 88 "nws/convex.nw"
void
convex::addPoint( const mathvec< double >& point )
{
    this->pts.insert( this->pts.end(), point );
};
#line 112 "nws/convex.nw"
class PlaneInfo {
public:
    PlaneInfo(
	const mathvec< double >& nn,
	const mathvec< double >& pp,
	double oo,
	std::vector< mathvec< double > >& pts
    ) : normal( nn ), point( pp ), offset( oo )
    {
	this->mask.resize( pts.size() );
	for ( unsigned int ii=0; ii < pts.size(); ++ii ) {
	    double vv = ( this->normal ^ pts[ ii ] ) - this->offset;
	    this->mask[ ii ] = ( ::fabs( vv ) < 0.000001 );
	}
    };

public:
    mathvec< double > normal;
    mathvec< double > point;
    double offset;
    std::vector< bool > mask;
    bool exceeded;
};

void
convex::calculateBoundingPlanes( bool dbg )
{
    std::list< PlaneInfo > planes;

    
#line 173 "nws/convex.nw"
unsigned int nn = this->pts.size();
if ( nn == 0 ) {
    return;
}

unsigned int rr = this->pts[ 0 ].size();
for ( unsigned int ii=1; ii < nn; ++ii ) {
    unsigned int len = this->pts[ ii ].size();
    if ( len > rr ) {
	rr = len;
    }
}

if ( rr > nn ) {
    std::cerr << "CONVEX: Not enough points to form a plane" << std::endl;
    return;
}

for ( unsigned int ii=0; ii < nn; ++ii ) {
    this->pts[ ii ].resize( rr, 0.0 );
}

#line 143 "nws/convex.nw"
    
#line 206 "nws/convex.nw"
double radius2 = 0.0;
for ( unsigned int kk=0; kk < nn; ++kk ) {
    const mathvec< double >& cur( this->pts[ kk ] );
    double r2 = cur ^ cur;
    if ( r2 > radius2 ) {
	radius2 = r2;
    }
}
#line 144 "nws/convex.nw"
    
#line 216 "nws/convex.nw"
cylinder* sphere = new cylinder( ::sqrt( radius2 )+0.01, this->myScene );
this->objs.insert( this->objs.begin(), sphere );

#line 146 "nws/convex.nw"
    
#line 197 "nws/convex.nw"
mathvec< unsigned int > choice;
choice.resize( rr );
for ( unsigned int ii=0; ii < rr; ++ii ) {
    unsigned int jj = rr - ii - 1;
    choice[ ii ] = jj;
}

#line 148 "nws/convex.nw"
    
#line 221 "nws/convex.nw"
do {
    
#line 231 "nws/convex.nw"
basevec< mathvec< double > > matrix;
matrix.resize( rr-1 );
mathvec< double >& first( this->pts[ choice[ 0 ] ] );
for ( unsigned int ii=1; ii < rr; ++ii ) {
    mathvec< double >& cur( this->pts[ choice[ ii ] ] );
    matrix[ ii-1 ] = cur - first;
}

bool plane = ::orthogonalize( matrix, rr );
#line 223 "nws/convex.nw"
    if ( plane ) {
	
#line 243 "nws/convex.nw"
mathvec< double > norm( matrix[ rr-1 ] );
double dd = ( norm ^ first );
#line 225 "nws/convex.nw"
	
#line 255 "nws/convex.nw"
{
    bool unique = true;

    std::list< PlaneInfo >::iterator pit = planes.begin();
    while ( unique && pit != planes.end() ) {
	if ( ::fabs( pit->offset - dd ) < 0.000001 ) {
	    if ( ( pit->normal - norm ).magnitude() < 0.000001 ) {
		unique = false;
	    }
	}
	++pit;
    }

    if ( unique ) {
	planes.insert( planes.end(), PlaneInfo( norm, first, dd, this->pts ) );
    }
}
#line 249 "nws/convex.nw"
norm *= -1.0;
dd = ( norm ^ first );
#line 255 "nws/convex.nw"
{
    bool unique = true;

    std::list< PlaneInfo >::iterator pit = planes.begin();
    while ( unique && pit != planes.end() ) {
	if ( ::fabs( pit->offset - dd ) < 0.000001 ) {
	    if ( ( pit->normal - norm ).magnitude() < 0.000001 ) {
		unique = false;
	    }
	}
	++pit;
    }

    if ( unique ) {
	planes.insert( planes.end(), PlaneInfo( norm, first, dd, this->pts ) );
    }
}
#line 226 "nws/convex.nw"
    }
} while ( planes.size() == 0 && this->incrementChoice( choice, nn ) );

#line 150 "nws/convex.nw"
    if ( planes.size() > 0 ) {
	for ( unsigned int kk=rr; kk < nn; ++kk ) {
	    const mathvec< double >& cur( this->pts[ kk ] );
	    
#line 275 "nws/convex.nw"
std::vector< mathvec< double > > pool;
std::vector< bool > pointMask( kk, false );
std::list< PlaneInfo >::iterator pit = planes.begin();
while ( pit != planes.end() ) {
    double diff = ( pit->normal ^ cur ) - pit->offset;
    pit->exceeded = ( diff > 0.000001 );
    if ( pit->exceeded ) {
	for ( unsigned int ii=0; ii < kk; ++ii ) {
	    if ( pit->mask[ ii ] == true && pointMask[ ii ] == false ) {
		pool.insert( pool.end(), this->pts[ ii ] );
		pointMask[ ii ] = true;
	    }
	}
    }
    std::list< PlaneInfo >::iterator last = pit++;
    if ( last->exceeded ) {
	planes.erase( last );
    }
}
#line 154 "nws/convex.nw"
	    if ( pool.size() >= rr-1 ) {
		
#line 311 "nws/convex.nw"
mathvec< unsigned int > choice;
choice.resize( rr-1 );
for ( unsigned int ii=0; ii < rr-1; ++ii ) {
    unsigned int jj = rr - ii - 2;
    choice[ ii ] = jj;
}
#line 298 "nws/convex.nw"
do {
    
#line 320 "nws/convex.nw"
basevec< mathvec< double > > matrix;
matrix.resize( rr-1 );
mathvec< double >& first( this->pts[ kk ] );
for ( unsigned int ii=0; ii < rr-1; ++ii ) {
    mathvec< double >& cur( pool[ choice[ ii ] ] );
    matrix[ ii ] = cur - first;
}

bool plane = ::orthogonalize( matrix, rr );
#line 300 "nws/convex.nw"
    if ( plane ) {
	
#line 243 "nws/convex.nw"
mathvec< double > norm( matrix[ rr-1 ] );
double dd = ( norm ^ first );
#line 302 "nws/convex.nw"
	
#line 332 "nws/convex.nw"
bool bounding = true;
unsigned int above = 0;
unsigned int below = 0;
unsigned int ps = pool.size();
for ( unsigned int ii=0; bounding && ii < ps; ++ii ) {
    double vv = ( norm ^ pool[ ii ] ) - dd;
    if ( vv > 0.000001 ) {
	++above;
    } else if ( vv < -0.000001 ) {
	++below;
    }
    bounding = ( above == 0 || below == 0 );
}
if ( bounding && above > 0 ) {
    norm *= -1.0;
    dd = ( norm ^ first );
}
#line 303 "nws/convex.nw"
	if ( bounding ) {
	    
#line 255 "nws/convex.nw"
{
    bool unique = true;

    std::list< PlaneInfo >::iterator pit = planes.begin();
    while ( unique && pit != planes.end() ) {
	if ( ::fabs( pit->offset - dd ) < 0.000001 ) {
	    if ( ( pit->normal - norm ).magnitude() < 0.000001 ) {
		unique = false;
	    }
	}
	++pit;
    }

    if ( unique ) {
	planes.insert( planes.end(), PlaneInfo( norm, first, dd, this->pts ) );
    }
}
#line 305 "nws/convex.nw"
	}
    }
} while ( this->incrementChoice( choice, pool.size() ) );
#line 156 "nws/convex.nw"
	    }
	}
	
#line 352 "nws/convex.nw"
std::list< PlaneInfo >::iterator pit = planes.begin();
while ( pit != planes.end() ) {
    halfspace* hh = new halfspace( pit->point, pit->normal );
    this->objs.insert( this->objs.end(), hh );
    ++pit;
}
#line 159 "nws/convex.nw"
    }

    if ( dbg ) {
	std::cerr << "(" << planes.size() << ") faces" << std::endl;
	std::list< PlaneInfo >::const_iterator it = planes.begin();
	while ( it != planes.end() ) {
	    std::cerr << "    " << it->normal << std::endl;
	    ++it;
	}
    }
}
#line 367 "nws/convex.nw"
bool
convex::incrementChoice( mathvec< unsigned int >& choice, unsigned int nn )
{
    unsigned int dims = choice.size();
    bool ret = false;

    for ( unsigned int ii=0; !ret && ii < dims; ++ii ) {
	if ( ++choice[ ii ] < nn - ii  ) {
	    for ( unsigned int jj=0; jj < ii; ++jj ) {
		unsigned int kk = ii - jj - 1;
		choice[ kk ] = choice[ ii ] + jj + 1;
	    }
	    ret = true;
	}
    }

    return ret;
}
#line 400 "nws/convex.nw"
bool
convex::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    bool ret = this->csginter::intersect( direction, start, hit );
    if ( ret && this->colorDistance > 0.000001 ) {
	
#line 418 "nws/convex.nw"
    const mathvec< double >& loc = hit.getPosition();
    mathvec< double > rp = loc - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    unsigned int clen = this->center.size();
    unsigned int cmax = clen;
    for ( unsigned int kk=0; kk < clen; ++kk ) {
	if ( this->specialColors[ kk ] != 0 ) {
	    cmax = clen - kk;
	    break;
	}
    }
    unsigned int len = this->specialColors.size();
    unsigned int hitCount = 0;
    unsigned int cindex = 0;
    std::vector< object* >::const_iterator jj = this->objs.begin();
    for ( ++jj; hitCount < cmax && jj != this->objs.end(); ++jj ) {
	halfspace* hh = (halfspace*)*jj;
	double dd = ::fabs( hh->distance( rp ) );
	if ( dd <= this->colorDistance ) {
	    cindex = clen - ++hitCount;
	}
    }
    if ( cindex < len ) {
	color* cc = this->specialColors[ cindex ];
	if ( cc != 0 ) {
	    hit.recolor( *cc );
	}
    }
#line 409 "nws/convex.nw"
    }
    return ret;
}
#line 467 "nws/convex.nw"
bool
convex::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    bool ret = this->csginter::intersect( direction, start, hits );
    if ( ret && this->colorDistance > 0.000001 ) {
	std::vector< intersection* >::iterator it;
	for ( it = hits.begin(); it != hits.end(); ++it ) {
	    intersection& hit( *(*it) );
	    
#line 418 "nws/convex.nw"
    const mathvec< double >& loc = hit.getPosition();
    mathvec< double > rp = loc - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    unsigned int clen = this->center.size();
    unsigned int cmax = clen;
    for ( unsigned int kk=0; kk < clen; ++kk ) {
	if ( this->specialColors[ kk ] != 0 ) {
	    cmax = clen - kk;
	    break;
	}
    }
    unsigned int len = this->specialColors.size();
    unsigned int hitCount = 0;
    unsigned int cindex = 0;
    std::vector< object* >::const_iterator jj = this->objs.begin();
    for ( ++jj; hitCount < cmax && jj != this->objs.end(); ++jj ) {
	halfspace* hh = (halfspace*)*jj;
	double dd = ::fabs( hh->distance( rp ) );
	if ( dd <= this->colorDistance ) {
	    cindex = clen - ++hitCount;
	}
    }
    if ( cindex < len ) {
	color* cc = this->specialColors[ cindex ];
	if ( cc != 0 ) {
	    hit.recolor( *cc );
	}
    }
#line 479 "nws/convex.nw"
	}
    }
    return ret;
}
#line 493 "nws/convex.nw"
std::istream&
operator >> ( std::istream& stream, convex& hh )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	
#line 533 "nws/convex.nw"
if ( tok == "colorDistance" ) {
    stream >> hh.colorDistance >> semicolon;
} else if ( tok == "specialColor" ) {
    unsigned int ii;
    color* nn = new color( hh.myScene );
    stream >> ii >> *nn;
    if ( ii >= hh.specialColors.size() ) {
	hh.specialColors.resize( ii+1, 0 );
    }
    color* oo = hh.specialColors[ ii ];
    if ( oo != 0 ) {
	delete oo;
    }
    hh.specialColors[ ii ] = nn;
}
#line 509 "nws/convex.nw"
	else if ( tok == "points" ) {
	    basevec< mathvec< double > > pts;
	    stream >> pts >> semicolon;
	    for ( unsigned int ii=0; ii < pts.size(); ++ii ) {
		hh.addPoint( pts[ii] );
	    }
	} else if ( tok == "point" ) {
	    mathvec< double > point;
	    stream >> point >> semicolon;
	    hh.addPoint( point );
	} else if ( tok == "base" ) {
	    stream >> (object&)hh >> semicolon;
	} else {
	    in.ignore( "CONVEX", tok.c_str() );
	}
    }

    hh.calculateBoundingPlanes();

    return stream;
}
