#line 324 "nws/coxeter.nw"
#include <string>
#include <sstream>
#include <math.h>
#include "input.h"
#include "halfspace.h"
#include "inter.h"
#include "convex.h"
#include "coxeter.h"

const char* coxeter::id = "NKVERSION: coxeter( 2.1.2007.04.07 )"; 

#line 54 "nws/coxeter.nw"
coxeter::coxeter( const scene* ss ) : convex( ss )
{
}
#line 67 "nws/coxeter.nw"
coxeter::~coxeter( void ) {
}
#line 85 "nws/coxeter.nw"
std::istream&
operator >> ( std::istream& stream, coxeter& hh )
{
    input in( stream );

    basevec< mathvec< unsigned int > > imat;
    unsigned int limit = 1024;
    bool debug = false;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	
#line 533 "nws/convex.nw"
if ( tok == "colorDistance" ) {
    stream >> hh.colorDistance >> semicolon;
} else if ( tok == "specialColor" ) {
    unsigned int ii;
    color* nn = new color( hh.myScene );
    stream >> ii >> *nn;
    if ( ii >= hh.specialColors.size() ) {
	hh.specialColors.resize( ii+1, 0 );
    }
    color* oo = hh.specialColors[ ii ];
    if ( oo != 0 ) {
	delete oo;
    }
    hh.specialColors[ ii ] = nn;
}
#line 105 "nws/coxeter.nw"
	else if ( tok == "matrix" || tok == "incidence" ) {
	    stream >> imat >> semicolon;
	} else if ( tok == "limit" ) {
	    stream >> limit >> semicolon;
	} else if ( tok == "debug" ) {
	    debug = true;
	    stream >> semicolon;
	} else if ( tok == "base" ) {
	    stream >> (object&)hh >> semicolon;
	} else {
	    std::cerr << "COXETER: " << tok << std::endl;
	}
    }

    if ( imat.size() > 0 ) {
	unsigned int nn = imat.size();
	
#line 132 "nws/coxeter.nw"
for ( unsigned int jj=0; jj < nn; ++jj ) {
    imat[jj].resize( nn, 0 );
}
#line 122 "nws/coxeter.nw"
	
#line 138 "nws/coxeter.nw"
basevec< mathvec< double > > dd;
dd.resize( nn );

for ( unsigned int jj=0; jj < nn; ++jj ) {
    mathvec< double >& row( dd[ jj ] );

    row.resize( nn, 0.0 );
    row *= 0.0;
    row[ jj ] = 1.0;
}

for ( unsigned int jj=0; jj < (nn-1); ++jj ) {
    for ( unsigned int ii=jj+1; ii < nn; ++ii ) {
	unsigned int mm = imat[ ii ][ jj ];
	unsigned int m2 = imat[ jj ][ ii ];
	if ( m2 > mm ) {
	    mm = m2;
	}
	if ( mm < 2 ) {
	    mm = 2;
	}
	double cc = ::cos( M_PI / (double)( mm ) );
	dd[ ii ][ jj ] = cc;
	dd[ jj ][ ii ] = cc;
    }
}
#line 123 "nws/coxeter.nw"
	
#line 167 "nws/coxeter.nw"
basevec< mathvec< double > > aa;
aa.resize( nn );

aa[ 0 ].resize( nn, 0.0 );
aa[ 0 ] *= 0.0;
aa[ 0 ][ 0 ] = 1.0;

for ( unsigned int kk=1; kk < nn; ++kk ) {
    mathvec< double >& cur( aa[ kk ] );
    double ll = 0.0;

    cur.resize( nn, 0.0 );
    cur *= 0.0;

    for ( unsigned int jj=0; jj < kk; ++jj ) {
	const mathvec< double >& versus( aa[ jj ] );
	double ss = 0.0;
	for ( unsigned int ii=0; ii < jj; ++ii ) {
	    ss += cur[ ii ] * versus[ ii ];
	}
	cur[ jj ] = ( dd[ jj ][ kk ] - ss ) / versus[ jj ];
	ll += cur[ jj ] * cur[ jj ];
    }

    cur[ kk ] = ::sqrt( 1.0 - ll );
}
#line 124 "nws/coxeter.nw"
	
#line 196 "nws/coxeter.nw"
unsigned int oldSize = 0;

basevec< mathvec< double > > xx;

xx.resize( xx.size() + 1 );
xx[ 0 ].resize( nn );

#line 235 "nws/coxeter.nw"
mathvec< double > bb( nn );
for ( unsigned int ii=0; ii < nn; ++ii ) {
    bb[ ii ] = (double)imat[ ii ][ ii ];
}

if ( ( bb ^ bb ) < 0.5 ) {
    bb[ 0 ] = 1.0;
}

xx[ 0 ] = bb.solve( aa );
if ( ( xx[ 0 ] ^ xx[ 0 ] ) < 0.5 ) {
    if ( debug ) {
	std::cerr << "bad solution: " << xx[0] << std::endl;
    }
    xx[ 0 ] = aa[ 0 ];
} else {
    xx[ 0 ].normalize();
    if ( debug ) {
	std::cerr << aa << " * " << xx[0] << " = " << bb << std::endl;
    }
}

#line 205 "nws/coxeter.nw"
unsigned int newSize = xx.size();

while ( newSize > oldSize && newSize < limit ) {
    unsigned int cntr = 0;

    for ( unsigned int jj=0; jj < nn; ++jj ) {
	for ( unsigned int ii=oldSize; ii < newSize; ++ii ) {
	    mathvec< double > vv;
	    
#line 259 "nws/coxeter.nw"
vv = aa[ jj ];
vv *= -2.0 * ( aa[ jj ] ^ xx[ ii ] );
vv += xx[ ii ];
#line 214 "nws/coxeter.nw"
	    
#line 265 "nws/coxeter.nw"
bool unique = true;
for ( unsigned int kk=0; unique && kk < newSize+cntr; ++kk ) {
    double dd = ( vv - xx[ kk ] ).magnitude();
    unique = dd > 0.000001;
}
if ( unique ) {
    xx.resize( newSize + cntr + 1 );
    xx[ newSize + cntr ] = vv;
    ++cntr;
}
#line 215 "nws/coxeter.nw"
	}
    }

    oldSize = newSize;
    newSize += cntr;
}

if ( debug ) {
    std::cerr << "(" << xx.size() << ") vertexes" << std::endl;
    std::cerr << xx << std::endl;
}

for ( unsigned int jj=0; jj < xx.size(); ++jj ) {
    hh.addPoint( xx[ jj ] );
}

hh.calculateBoundingPlanes( debug );
#line 125 "nws/coxeter.nw"
    }

    return stream;
}
