#line 553 "nws/cylinder.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "cylinder.h"

const char* cylinder::id = "NKVERSION: cylinder( 2.1.2007.05.17 )"; 

#line 65 "nws/cylinder.nw"
cylinder::cylinder( const scene* ss ) : object( ss )
{
}
#line 77 "nws/cylinder.nw"
cylinder::cylinder( double radius, const scene* ss ) : object( ss )
{
    this->defaultScale = radius;
    this->roundDimensions = ~0U;
}
#line 92 "nws/cylinder.nw"
cylinder::~cylinder( void ) {
};
#line 105 "nws/cylinder.nw"
void
cylinder::setDimensions( unsigned int dd )
{
    this->sphereMask.resize( dd );

    if ( this->roundDimensions == 1 ) {
	this->roundDimensions = 0;
    }

    if ( this->roundDimensions > dd ) {
	this->roundDimensions = dd;
    }

    for ( unsigned int ii=0; ii < this->roundDimensions; ++ii ) {
	this->sphereMask[ ii ] = 1.0;
    }

    for ( unsigned int ii=this->roundDimensions; ii < dd; ++ii ) {
	this->sphereMask[ ii ] = 0.0;
    }

    this->object::setDimensions( dd );
}
#line 144 "nws/cylinder.nw"
bool
cylinder::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    
#line 182 "nws/cylinder.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

double t1 = -MAXFLOAT;
double t2 = MAXFLOAT;

mathvec< double > n1;
mathvec< double > n2;

bool hitSphere = this->sphereIntersect( rd, rp, &t1, &t2 );
bool hitCube = hitSphere && this->cubeIntersect( rd, rp, &t1, &t2, n1, n2 );

#line 152 "nws/cylinder.nw"
    if ( !hitSphere || !hitCube ) {
	if ( this->_inside( rp ) ) {
	    this->forgeHit( direction, start, hit );
	    return true;
	} else {
	    return false;
	}
    }

    if ( t1 >= 0.000001 ) {
	mathvec< double >& nn( n1 );
	double tt = t1;
	double normalFactor = 1.0;
	
#line 202 "nws/cylinder.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( nn.size() == 0 ) {
    nn = spot * this->sphereMask * normalFactor;
    nn.normalize();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 166 "nws/cylinder.nw"
	hit.setOutside();
    } else if ( t2 >= 0.000001 ) {
	mathvec< double >& nn( n2 );
	double tt = t2;
	double normalFactor = -1.0;
	
#line 202 "nws/cylinder.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( nn.size() == 0 ) {
    nn = spot * this->sphereMask * normalFactor;
    nn.normalize();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 172 "nws/cylinder.nw"
	hit.setInside();
    } else {
	return false;
    }

    return true;
}
#line 236 "nws/cylinder.nw"
bool
cylinder::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    
#line 182 "nws/cylinder.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

double t1 = -MAXFLOAT;
double t2 = MAXFLOAT;

mathvec< double > n1;
mathvec< double > n2;

bool hitSphere = this->sphereIntersect( rd, rp, &t1, &t2 );
bool hitCube = hitSphere && this->cubeIntersect( rd, rp, &t1, &t2, n1, n2 );

#line 244 "nws/cylinder.nw"
    hits.clear();

    if ( !hitSphere || !hitCube ) {
	if ( this->_inside( rp ) ) {
	    intersection* hh = new intersection();
	    this->forgeHit( direction, start, *hh );
	    hits.insert( hits.end(), hh );
	    return true;
	} else {
	    return false;
	}
    }

    if ( t1 >= 0.000001 ) {
	intersection* h1 = new intersection();
	intersection& hit( *h1 );
	mathvec< double >& nn( n1 );
	double tt = t1;
	double normalFactor = 1.0;
	
#line 202 "nws/cylinder.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( nn.size() == 0 ) {
    nn = spot * this->sphereMask * normalFactor;
    nn.normalize();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 264 "nws/cylinder.nw"
	hit.setOutside();
	hits.insert( hits.end(), h1 );
    } 

    if ( t2 >= 0.000001 ) {
	intersection* h2 = new intersection();
	intersection& hit( *h2 );
	mathvec< double >& nn( n2 );
	double tt = t2;
	double normalFactor = -1.0;
	
#line 202 "nws/cylinder.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( nn.size() == 0 ) {
    nn = spot * this->sphereMask * normalFactor;
    nn.normalize();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 275 "nws/cylinder.nw"
	hit.setInside();
	hits.insert( hits.end(), h2 );
    }

    return hits.size() > 0;
}
#line 293 "nws/cylinder.nw"
bool
cylinder::sphereIntersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    double* t1, double* t2
) const {
    if ( this->roundDimensions == 0 ) {
	return true;
    }

    mathvec< double > rds = direction * this->sphereMask;
    mathvec< double > rps = start * this->sphereMask;

    double aa = rds ^ rds;
    double cc = ( rps ^ rps ) - 1.0;

    if ( aa < 0.000001 ) {
	return ( cc < 0.000001 );
    }

    double bb = 2.0 * ( rds ^ rps );
    double discriminant = bb * bb - 4.0 * aa * cc;


    if ( discriminant < 0.000001 ) {
	return false;
    }

    double root = ::sqrt( discriminant );

    if ( aa > 0 ) {
	*t1 = ( -bb - root ) / ( 2.0 * aa );
	*t2 = ( -bb + root ) / ( 2.0 * aa );
    } else {
	*t1 = ( -bb + root ) / ( 2.0 * aa );
	*t2 = ( -bb - root ) / ( 2.0 * aa );
    }

    if ( *t2 < 0.000001 ) {
	return false;
    }

    return true;
}
#line 351 "nws/cylinder.nw"
bool
cylinder::cubeIntersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    double* t1, double* t2,
    mathvec< double >& n1, mathvec< double >& n2
) const {
    unsigned int dims = this->center.size();

    if ( this->roundDimensions == dims ) {
	return true;
    }

    unsigned int n1dim = dims;
    double n1mag = 1.0;
    unsigned int n2dim = dims;
    double n2mag = 1.0;

    for ( unsigned int ii=this->roundDimensions; ii < dims; ++ii ) {
	double t1c;
	double t2c;
	double nn = direction[ ii ];
	double ss = start[ ii ];

	if ( ::fabs( nn ) < 0.000001 ) {
	    if ( ss < -1.0 || ss > 1.0 ) {
		return false;
	    } else {
		continue;
	    }
	}

	t1c = ( -ss + 1.0 ) / nn;
	t2c = ( -ss - 1.0 ) / nn;

	if ( t1c < 0.000001 && t2c < 0.000001 ) {
	    return false;
	}

	double mag = 1.0;

	if ( t1c > t2c ) {
	    double tmp = t1c;
	    t1c = t2c;
	    t2c = tmp;
	    mag = -1.0;
	}

	if ( t2c < *t1 || t1c > *t2  ) {
	    return false;
	}

	if ( t1c > *t1 ) {
	    *t1 = t1c;
	    n1dim = ii;
	    n1mag = mag;
	}
	if ( t2c < *t2 ) {
	    *t2 = t2c;
	    n2dim = ii;
	    n2mag = mag;
	}
    }

    if ( n1dim < dims ) {
	n1.resize( dims, 0.0 );
	n1[ n1dim ] = n1mag;
    }
    if ( n2dim < dims ) {
	n2.resize( dims, 0.0 );
	n2[ n2dim ] = n2mag;
    }

    return true;
}
#line 436 "nws/cylinder.nw"
bool
cylinder::inside( const mathvec< double >& start ) const
{
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
cylinder::_inside( const mathvec< double >& rp ) const
{
    unsigned int dims = this->center.size();
    for ( unsigned int ii=this->roundDimensions; ii < dims; ++ii ) {
	if ( ::fabs( rp[ ii ] ) > 1.0 ) {
	    return false;
	}
    }

    double sum = 0.0;
    for ( unsigned int ii=0; ii < this->roundDimensions; ++ii ) {
	sum += rp[ ii ] * rp[ ii ];
    }

    return sum <= 1.0;
}
#line 473 "nws/cylinder.nw"
std::istream&
operator >> ( std::istream& stream, cylinder& cc )
{
    input in( stream );

    cc.roundDimensions = 0;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "roundDimensions" ) {
	    stream >> cc.roundDimensions >> semicolon;
	} else if ( tok == "base" ) {
	    stream >> (object&)cc >> semicolon;
	} else {
	    in.ignore( "CYLINDER", tok.c_str() );
	}
    }

    return stream;
}
