#line 411 "nws/expr.nw"
#include <string>
#include <sstream>
#include <math.h>
#include "input.h"
#include "expr.h"
#include "portal.h"

const char* expr::id = "NKVERSION: expr( 2.1.2007.03.21 )"; 

#line 258 "nws/expr-binary.nw"
class exprBinary : public expr {
public:
    
#line 39 "nws/expr-binary.nw"
exprBinary( double (*_func)( double, double ) )
    : func( _func ), left( 0 ), right( 0 )
{
};
#line 48 "nws/expr-binary.nw"
~exprBinary( void )
{
    delete this->right;
    delete this->left;
};
#line 60 "nws/expr-binary.nw"
virtual bool isConstant( void ) const
{
    return this->left->isConstant() && this->right->isConstant();
};
#line 66 "nws/expr-binary.nw"
virtual bool usesPosition( void ) const
{
    return this->left->usesPosition() || this->right->usesPosition();
};
#line 72 "nws/expr-binary.nw"
virtual bool usesNormal( void ) const
{
    return this->left->usesNormal() || this->right->usesNormal();
};
#line 78 "nws/expr-binary.nw"
virtual bool usesDirection( void ) const
{
    return this->left->usesDirection() || this->right->usesDirection();
};
#line 90 "nws/expr-binary.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    return (*this->func)(
	this->left->evaluate( pos, norm, dir ),
	this->right->evaluate( pos, norm, dir )
    );
}
#line 107 "nws/expr-binary.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->left;
	this->left = expr::read( in, portals );
    delete this->right;
	this->right = expr::read( in, portals );
    return in;
};
#line 261 "nws/expr-binary.nw"
protected:
    
#line 30 "nws/expr-binary.nw"
double (*func)( double, double );
expr* left;
expr* right;
#line 263 "nws/expr-binary.nw"
public:
    static const char* id;
};

const char* exprBinary::id = "NKVERSION: exprBinary( 2.1.2007.03.21 )"; 
#line 216 "nws/expr-unary.nw"
class exprUnary : public expr {
public:
    
#line 52 "nws/expr-unary.nw"
exprUnary( double (*_func)( double ) )
    : func( _func ), operand( 0 )
{
};
#line 61 "nws/expr-unary.nw"
~exprUnary( void )
{
    delete this->operand;
};
#line 71 "nws/expr-unary.nw"
virtual bool isConstant( void ) const
{
    return this->operand->isConstant();
};
#line 77 "nws/expr-unary.nw"
virtual bool usesPosition( void ) const
{
    return this->operand->usesPosition();
};
#line 83 "nws/expr-unary.nw"
virtual bool usesNormal( void ) const
{
    return this->operand->usesNormal();
};
#line 89 "nws/expr-unary.nw"
virtual bool usesDirection( void ) const
{
    return this->operand->usesDirection();
};
#line 100 "nws/expr-unary.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    return (*this->func)( this->operand->evaluate( pos, norm, dir ) );
}
#line 114 "nws/expr-unary.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->operand;
	this->operand = expr::read( in, portals );
    return in;
};
#line 219 "nws/expr-unary.nw"
protected:
    
#line 45 "nws/expr-unary.nw"
double (*func)( double );
expr* operand;
#line 221 "nws/expr-unary.nw"
public:
    static const char* id;
};

const char* exprUnary::id = "NKVERSION: exprUnary( 2.1.2007.03.21 )"; 
#line 143 "nws/expr-const.nw"
class exprConst : public expr {
public:
    
#line 25 "nws/expr-const.nw"
exprConst( double _value ) : value( _value ) {};
#line 32 "nws/expr-const.nw"
virtual bool isConstant( void ) const
{
    return true;
};
#line 38 "nws/expr-const.nw"
virtual bool usesPosition( void ) const
{
    return false;
};
#line 44 "nws/expr-const.nw"
virtual bool usesNormal( void ) const
{
    return false;
};
#line 50 "nws/expr-const.nw"
virtual bool usesDirection( void ) const
{
    return false;
};
#line 60 "nws/expr-const.nw"
virtual double evaluate(
	const mathvec<double>& /*position*/,
	const mathvec<double>& /*normal*/,
	const mathvec<double>& /*direction*/
    )
{
    return this->value;
}
#line 76 "nws/expr-const.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& /*portals*/
) {
    return in;
};
#line 146 "nws/expr-const.nw"
protected:
    
#line 18 "nws/expr-const.nw"
double value;
#line 148 "nws/expr-const.nw"
public:
    static const char* id;
};

const char* exprConst::id = "NKVERSION: exprConst( 2.1.2007.03.21 )"; 
#line 108 "nws/expr-pos.nw"
class exprPosition : public expr {
public:
    
#line 25 "nws/expr-pos.nw"
exprPosition( void )
    : operand( 0 )
{
};
#line 34 "nws/expr-pos.nw"
~exprPosition( void )
{
    delete this->operand;
};
#line 45 "nws/expr-pos.nw"
virtual bool isConstant( void ) const
{
    return false;
};
#line 51 "nws/expr-pos.nw"
virtual bool usesPosition( void ) const
{
    return true;
};
#line 57 "nws/expr-pos.nw"
virtual bool usesNormal( void ) const
{
    return this->operand->usesNormal();
};
#line 63 "nws/expr-pos.nw"
virtual bool usesDirection( void ) const
{
    return this->operand->usesDirection();
};
#line 75 "nws/expr-pos.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    double op = this->operand->evaluate( pos, norm, dir );
    unsigned int index = (unsigned int)( op + 0.5 );
    return pos[ index ];
}
#line 91 "nws/expr-pos.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->operand;
	this->operand = expr::read( in, portals );
    return in;
};
#line 111 "nws/expr-pos.nw"
protected:
    
#line 19 "nws/expr-pos.nw"
expr* operand;
#line 113 "nws/expr-pos.nw"
public:
    static const char* id;
};

const char* exprPosition::id = "NKVERSION: exprPosition( 2.1.2007.03.21 )"; 
#line 112 "nws/expr-dir.nw"
class exprDirection : public expr {
public:
    
#line 29 "nws/expr-dir.nw"
exprDirection( void )
    : operand( 0 )
{
};
#line 38 "nws/expr-dir.nw"
~exprDirection( void )
{
    delete this->operand;
};
#line 49 "nws/expr-dir.nw"
virtual bool isConstant( void ) const
{
    return false;
};
#line 55 "nws/expr-dir.nw"
virtual bool usesPosition( void ) const
{
    return this->operand->usesPosition();
};
#line 61 "nws/expr-dir.nw"
virtual bool usesNormal( void ) const
{
    return this->operand->usesNormal();
};
#line 67 "nws/expr-dir.nw"
virtual bool usesDirection( void ) const
{
    return true;
};
#line 79 "nws/expr-dir.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    double op = this->operand->evaluate( pos, norm, dir );
    unsigned int index = (unsigned int)( op + 0.5 );
    return dir[ index ];
}
#line 95 "nws/expr-dir.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->operand;
	this->operand = expr::read( in, portals );
    return in;
};
#line 115 "nws/expr-dir.nw"
protected:
    
#line 23 "nws/expr-dir.nw"
expr* operand;
#line 117 "nws/expr-dir.nw"
public:
    static const char* id;
};

const char* exprDirection::id = "NKVERSION: exprDirection( 2.1.2007.03.21 )"; 
#line 111 "nws/expr-norm.nw"
class exprNormal : public expr {
public:
    
#line 28 "nws/expr-norm.nw"
exprNormal( void )
    : operand( 0 )
{
};
#line 37 "nws/expr-norm.nw"
~exprNormal( void )
{
    delete this->operand;
};
#line 48 "nws/expr-norm.nw"
virtual bool isConstant( void ) const
{
    return false;
};
#line 54 "nws/expr-norm.nw"
virtual bool usesPosition( void ) const
{
    return this->operand->usesPosition();
};
#line 60 "nws/expr-norm.nw"
virtual bool usesNormal( void ) const
{
    return true;
};
#line 66 "nws/expr-norm.nw"
virtual bool usesDirection( void ) const
{
    return this->operand->usesDirection();
};
#line 78 "nws/expr-norm.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    double op = this->operand->evaluate( pos, norm, dir );
    unsigned int index = (unsigned int)( op + 0.5 );
    return norm[ index ];
}
#line 94 "nws/expr-norm.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->operand;
	this->operand = expr::read( in, portals );
    return in;
};
#line 114 "nws/expr-norm.nw"
protected:
    
#line 22 "nws/expr-norm.nw"
expr* operand;
#line 116 "nws/expr-norm.nw"
public:
    static const char* id;
};

const char* exprNormal::id = "NKVERSION: exprNormal( 2.1.2007.03.21 )"; 
#line 118 "nws/expr-ref.nw"
class exprReference : public expr {
public:
    
#line 34 "nws/expr-ref.nw"
exprReference( const mathvec< double >& pp )
    : pixel( pp ), operand( 0 )
{
};
#line 44 "nws/expr-ref.nw"
~exprReference( void )
{
    delete this->operand;
};
#line 55 "nws/expr-ref.nw"
virtual bool isConstant( void ) const
{
    return false;
};
#line 61 "nws/expr-ref.nw"
virtual bool usesPosition( void ) const
{
    return true;
};
#line 67 "nws/expr-ref.nw"
virtual bool usesNormal( void ) const
{
    return true;
};
#line 73 "nws/expr-ref.nw"
virtual bool usesDirection( void ) const
{
    return true;
};
#line 85 "nws/expr-ref.nw"
virtual double evaluate(
	const mathvec<double>& pos,
	const mathvec<double>& norm,
	const mathvec<double>& dir
    )
{
    double op = this->operand->evaluate( pos, norm, dir );
    unsigned int index = (unsigned int)( op + 0.5 );
    return this->pixel[ index ];
}
#line 101 "nws/expr-ref.nw"
virtual std::istream& get(
    std::istream& in,
    const std::map< std::string, class portal* >& portals
) {
    delete this->operand;
	this->operand = expr::read( in, portals );
    return in;
};
#line 121 "nws/expr-ref.nw"
protected:
    
#line 26 "nws/expr-ref.nw"
const mathvec< double >& pixel;
expr* operand;
#line 123 "nws/expr-ref.nw"
public:
    static const char* id;
};

const char* exprReference::id = "NKVERSION: exprReference( 2.1.2007.03.21 )"; 

#line 125 "nws/expr-binary.nw"
class BinaryExpressionTable {
protected:
    typedef double (*FunctionType)( double, double );
    typedef std::map< std::string, FunctionType > TableType;
    typedef TableType::const_iterator const_iterator;
    TableType table;
public:
    BinaryExpressionTable( void );
    exprBinary* operator [] ( const std::string& name ) const;

    static double add( double aa, double bb );
    static double sub( double aa, double bb );
    static double mul( double aa, double bb );
    static double div( double aa, double bb );
    static double mod( double aa, double bb );
    static double idiv( double aa, double bb );
    static double min( double aa, double bb );
    static double max( double aa, double bb );
public:
    static const char* id;
};

const char* BinaryExpressionTable::id
    = "NKVERSION: BinaryExpressionTable( 2.1.2007.03.21 )"; 
#line 155 "nws/expr-binary.nw"
BinaryExpressionTable::BinaryExpressionTable( void )
{
    this->table[ "+" ] = BinaryExpressionTable::add;
    this->table[ "-" ] = BinaryExpressionTable::sub;
    this->table[ "*" ] = BinaryExpressionTable::mul;
    this->table[ "/" ] = BinaryExpressionTable::div;
    this->table[ "%" ] = BinaryExpressionTable::mod;
    this->table[ "div" ] = BinaryExpressionTable::idiv;
    this->table[ "atan2" ] = (double (*)( double, double ))::atan2;
    this->table[ "^" ] = (double (*)( double, double ))::pow;
    this->table[ "min" ] = BinaryExpressionTable::min;
    this->table[ "max" ] = BinaryExpressionTable::max;
}
#line 174 "nws/expr-binary.nw"
exprBinary*
BinaryExpressionTable::operator [] ( const std::string& name ) const
{
    const_iterator it = this->table.find( name );
    if ( it != this->table.end() ) {
	return new exprBinary( it->second );
    } else {
	return 0;
    }
}
#line 189 "nws/expr-binary.nw"
double BinaryExpressionTable::add( double aa, double bb ) {
    return aa + bb;
}

double BinaryExpressionTable::sub( double aa, double bb ) {
    return aa - bb;
}

double BinaryExpressionTable::mul( double aa, double bb ) {
    return aa * bb;
}

double BinaryExpressionTable::div( double aa, double bb ) {
    return aa / bb;
}
#line 213 "nws/expr-binary.nw"
double BinaryExpressionTable::mod( double aa, double bb ) {
    double ii;
    double rr = ::modf( aa / bb, &ii );

    return rr * bb;
}
#line 224 "nws/expr-binary.nw"
double BinaryExpressionTable::idiv( double aa, double bb ) {
    double ii;
    (void) ::modf( aa / bb, &ii );

    return ii * bb;
}
#line 234 "nws/expr-binary.nw"
double BinaryExpressionTable::min( double aa, double bb ) {
    if ( aa < bb ) {
	return aa;
    } else {
	return bb;
    }
}
double BinaryExpressionTable::max( double aa, double bb ) {
    if ( aa < bb ) {
	return bb;
    } else {
	return aa;
    }
}
#line 130 "nws/expr-unary.nw"
class UnaryExpressionTable {
protected:
    typedef double (*FunctionType)( double );
    typedef std::map< std::string, FunctionType > TableType;
    typedef TableType::const_iterator const_iterator;
    TableType table;
public:
    UnaryExpressionTable( void );
    exprUnary* operator [] ( const std::string& name ) const;

    static double sign( double );
public:
    static const char* id;
};
const char* UnaryExpressionTable::id
    = "NKVERSION: UnaryExpressionTable( 2.1.2007.03.21 )"; 
#line 153 "nws/expr-unary.nw"
UnaryExpressionTable::UnaryExpressionTable( void )
{
    this->table[ "sin" ] = (double (*)( double ))::sin;
    this->table[ "cos" ] = (double (*)( double ))::cos;
    this->table[ "tan" ] = (double (*)( double ))::tan;
    this->table[ "atan" ] = (double (*)( double ))::atan;
    this->table[ "sinh" ] = (double (*)( double ))::sinh;
    this->table[ "cosh" ] = (double (*)( double ))::cosh;
    this->table[ "tanh" ] = (double (*)( double ))::tanh;
    this->table[ "sqrt" ] = (double (*)( double ))::sqrt;
    this->table[ "cbrt" ] = (double (*)( double ))::cbrt;
    this->table[ "exp" ] = (double (*)( double ))::exp;
    this->table[ "log" ] = (double (*)( double ))::log;
    this->table[ "log10" ] = (double (*)( double ))::log10;
    this->table[ "fabs" ] = (double (*)( double ))::fabs;
    this->table[ "floor" ] = (double (*)( double ))::floor;
    this->table[ "ceil" ] = (double (*)( double ))::ceil;
    this->table[ "rint" ] = (double (*)( double ))::rint;
    this->table[ "round" ] = (double (*)( double ))::rint;
    this->table[ "sign" ] = UnaryExpressionTable::sign;
    this->table[ "gamma" ] = (double (*)( double ))::gamma;
    this->table[ "lgamma" ] = (double (*)( double ))::lgamma;
    this->table[ "erf" ] = (double (*)( double ))::erf;
}
#line 182 "nws/expr-unary.nw"
double UnaryExpressionTable::sign( double vv ) {
    if ( vv > 0.000001 ) {
	return  1.0;
    } else if ( vv < -0.000001 ) {
	return -1.0;
    } else {
	return  0.0;
    }
}
#line 197 "nws/expr-unary.nw"
exprUnary*
UnaryExpressionTable::operator [] ( const std::string& name ) const
{
    const_iterator it = this->table.find( name );
    if ( it != this->table.end() ) {
	return new exprUnary( it->second );
    } else {
	return 0;
    }
}
#line 90 "nws/expr-const.nw"
class ConstantExpressionTable {
protected:
    typedef std::map< std::string, double > TableType;
    typedef TableType::const_iterator const_iterator;
    TableType table;
public:
    ConstantExpressionTable( void );
    exprConst* operator [] ( const std::string& name ) const;
public:
    static const char* id;
};

const char* ConstantExpressionTable::id
    = "NKVERSION: ConstantExpressionTable( 2.1.2007.03.21 )"; 
#line 110 "nws/expr-const.nw"
ConstantExpressionTable::ConstantExpressionTable( void )
{
    this->table[ "pi" ] = M_PI;
    this->table[ "2pi" ] = 2.0 * M_PI;
    this->table[ "e" ] = M_E;
    this->table[ "sqrt2" ] = M_SQRT2;
}
#line 123 "nws/expr-const.nw"
exprConst*
ConstantExpressionTable::operator [] ( const std::string& name ) const
{
    const_iterator it = this->table.find( name );
    if ( it != this->table.end() ) {
	return new exprConst( it->second );
    } else {
	return 0;
    }
}
#line 334 "nws/expr.nw"
class TableInitializer {
public:
    TableInitializer( void );
public:
    static const char* id;
};
const char* TableInitializer::id = "NKVERSION: TableInitializer( 2.1.2007.03.21 )"; 
#line 346 "nws/expr.nw"
TableInitializer::TableInitializer( void ) {
    std::map<std::string, class portal*> portals;
    expr* ee = expr::read( "+ sin pi 1", portals );
    if ( ee != 0 ) {
	delete ee;
    }
};
#line 359 "nws/expr.nw"
static TableInitializer __tableInitializer;

#line 52 "nws/expr.nw"
expr::~expr( void ) {}
#line 106 "nws/expr.nw"
expr*
expr::read(
    std::istream& stream,
    const std::map<std::string, class portal*>& portals
) {
    input in( stream );

    if ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return 0;
	}

	std::string tok( token );

	if ( tok == "pos" ) {
	    expr* nn = new exprPosition();
	    nn->get( stream, portals );
	    return nn;
	} else if ( tok == "dir" ) {
	    expr* nn = new exprDirection();
	    nn->get( stream, portals );
	    return nn;
	} else if ( tok == "norm" ) {
	    expr* nn = new exprNormal();
	    nn->get( stream, portals );
	    return nn;
	} else if ( tok == "@" ) {
	    
#line 152 "nws/expr.nw"
if ( in.get( token, sizeof(token) ) == 0 ) {
    return 0;
}

std::string name( token );

std::map< std::string, class portal* >::const_iterator pp
    = portals.find( name );
if ( pp == portals.end() ) {
    return 0;
}

exprReference* rret = new exprReference( pp->second->getPixel() );
rret->get( stream, portals );
return rret;
#line 136 "nws/expr.nw"
	}

	
#line 174 "nws/expr.nw"
static const BinaryExpressionTable binaries;
exprBinary* bret = binaries[ tok ];

if ( bret != 0 ) {
    bret->get( stream, portals );
    if ( bret->isConstant() ) {
	mathvec< double > dd;
	exprConst* cret = new exprConst( bret->evaluate( dd, dd, dd ) );
	delete bret;
	return cret;
    } else {
	return bret;
    }
}
#line 139 "nws/expr.nw"
	
#line 195 "nws/expr.nw"
static const UnaryExpressionTable unaries;
exprUnary* uret = unaries[ tok ];

if ( uret != 0 ) {
    uret->get( stream, portals );
    if ( uret->isConstant() ) {
	mathvec< double > dd;
	exprConst* cret = new exprConst( uret->evaluate( dd, dd, dd ) );
	delete uret;
	return cret;
    } else {
	return uret;
    }
}
#line 140 "nws/expr.nw"
	
#line 216 "nws/expr.nw"
static const ConstantExpressionTable constants;
exprConst* cret = constants[ tok ];

if ( cret != 0 ) {
    cret->get( stream, portals );
    return cret;
}
#line 229 "nws/expr.nw"
    double value;
    std::istringstream str( token );
    str >> value;

    if ( ! str.fail() ) {
	return new exprConst( value );
    }
#line 141 "nws/expr.nw"
    }

    return 0;
}
#line 243 "nws/expr.nw"
expr*
expr::read(
    const char* str,
    const std::map<std::string, class portal*>& portals
) {
    std::istringstream in( str );
    return read( in, portals );
}
#line 281 "nws/expr.nw"
void
expr::read(
    std::istream& in,
    basevec< expr* >& exprs,
    const std::map<std::string, class portal*>& portals
) {
    char bogus;
    unsigned int len;

    in.setf( std::ios::skipws );

    in >> bogus;		// "["
    in >> len;
    in >> bogus >> bogus;	// "]<"

    exprs.resize( len, (expr*)0 );

    for (unsigned ii=0; ii < len; ++ii) {
	exprs[ ii ] = expr::read( in, portals );

	if (ii < len-1) {
	    in >> bogus; // ","
	}
    }

    while (bogus != '>' && in.good() ) {
	in >> bogus; // ">"
    }
}
#line 314 "nws/expr.nw"
void
expr::read(
    const char* str,
    basevec< expr* >& exprs,
    const std::map<std::string, class portal*>& portals
) {
    std::istringstream in( str );
    expr::read( in, exprs, portals );
}
