#line 349 "nws/extrusion.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "extrusion.h"

const char* extrusion::id = "NKVERSION: extrusion( 2.1.2007.03.21 )"; 

#line 50 "nws/extrusion.nw"
extrusion::extrusion( const scene* ss )
    : object( ss ) 
#line 39 "nws/extrusion.nw"
, subDimensions( 0 ), obj( 0 )
#line 52 "nws/extrusion.nw"
{
}
#line 64 "nws/extrusion.nw"
extrusion::~extrusion( void ) {
    delete this->obj;
};
#line 78 "nws/extrusion.nw"
void
extrusion::setDimensions( unsigned int dd )
{
    if ( this->subDimensions == 0 || this->subDimensions > dd ) {
	this->subDimensions = dd;
    }
    if ( this->obj != 0 ) {
	this->obj->setDimensions( this->subDimensions );
    }
    this->object::setDimensions( dd );
}
#line 105 "nws/extrusion.nw"
bool
extrusion::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    if ( this->obj == 0 ) {
	return false;
    }
    
#line 127 "nws/extrusion.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;

mathvec< double > fullrd = rd;
rd.resize( this->subDimensions, 0.0 );

fullrd.normalize();
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

mathvec< double > fullrp = rp;
rp.resize( this->subDimensions, 0.0 );
#line 115 "nws/extrusion.nw"
    bool ret = false;
    if ( ( rd ^ rd ) > 0.000001 ) {
	ret = this->obj->intersect( rd, rp, hit );
	if ( ret ) {
	    
#line 146 "nws/extrusion.nw"
mathvec< double > hitPosition;
if ( hit.getDistance() < MAXFLOAT / 2.0 ) {
    hitPosition = fullrd;
    rd.resize( fullrd.size(), 0.0 );
    hitPosition *= hit.getDistance();
    double distFactor = ( rd ^ fullrd );
    if ( distFactor > 0.000001 ) {
	hitPosition /= distFactor;
    }
    hitPosition += fullrp;
    hit.setPosition( hitPosition );
} else {
    hitPosition = hit.getPosition();
    hitPosition.resize( this->center.size(), 0.0 );
    hit.setPosition( hitPosition );
}

hit.setDirection( fullrd );

mathvec< double > hitNormal = hit.getNormal();
hitNormal.resize( this->center.size(), 0.0 );
hit.setNormal( hitNormal );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hitPosition - start ).magnitude() );

if ( this->gotColor ) {
    hit.recolor( this->chroma );
}
#line 120 "nws/extrusion.nw"
	}
    }
    return ret;
}
#line 190 "nws/extrusion.nw"
bool
extrusion::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    if ( this->obj == 0 ) {
	return false;
    }
    
#line 127 "nws/extrusion.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;

mathvec< double > fullrd = rd;
rd.resize( this->subDimensions, 0.0 );

fullrd.normalize();
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

mathvec< double > fullrp = rp;
rp.resize( this->subDimensions, 0.0 );
#line 200 "nws/extrusion.nw"
    hits.clear();
    bool ret;

    if ( ( rd ^ rd ) > 0.000001 ) {
	ret = this->obj->intersect( rd, rp, hits );
	if ( ret ) {
	    for ( unsigned int ii=0; ii < hits.size(); ++ii ) {
		intersection& hit( *hits[ ii ] );
		
#line 146 "nws/extrusion.nw"
mathvec< double > hitPosition;
if ( hit.getDistance() < MAXFLOAT / 2.0 ) {
    hitPosition = fullrd;
    rd.resize( fullrd.size(), 0.0 );
    hitPosition *= hit.getDistance();
    double distFactor = ( rd ^ fullrd );
    if ( distFactor > 0.000001 ) {
	hitPosition /= distFactor;
    }
    hitPosition += fullrp;
    hit.setPosition( hitPosition );
} else {
    hitPosition = hit.getPosition();
    hitPosition.resize( this->center.size(), 0.0 );
    hit.setPosition( hitPosition );
}

hit.setDirection( fullrd );

mathvec< double > hitNormal = hit.getNormal();
hitNormal.resize( this->center.size(), 0.0 );
hit.setNormal( hitNormal );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hitPosition - start ).magnitude() );

if ( this->gotColor ) {
    hit.recolor( this->chroma );
}
#line 209 "nws/extrusion.nw"
	    }
	}
    } else {
	ret = this->obj->inside( rp );
	if ( ret ) {
	    intersection* nn = new intersection();
	    this->forgeHit( direction, start, *nn );
	    hits.insert( hits.end(), nn );
	}
    }

    return ret;
}
#line 232 "nws/extrusion.nw"
bool
extrusion::inside( const mathvec< double >& start ) const
{
    if ( this->obj == 0 ) {
	return false;
    }
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    rp.resize( this->subDimensions, 0.0 );

    return this->_inside( rp );
}

bool
extrusion::_inside( const mathvec< double >& rp ) const
{
    if ( this->obj == 0 ) {
	return false;
    }
    return this->obj->inside( rp );
}
#line 265 "nws/extrusion.nw"
std::istream&
operator >> ( std::istream& stream, extrusion& ee )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "dimensions" ) {
	    stream >> ee.subDimensions >> semicolon;
	} else if ( tok == "object" ) {
	    object* nn = object::read( stream, ee.myScene );
	    if ( nn != 0 ) {
		delete ee.obj;
		ee.obj = nn;
	    } else {
		std::cerr << "EXTRUSION failed to get an object!" << std::endl;
	    }
	} else if ( tok == "base" ) {
	    stream >> (object&)ee >> semicolon;
	} else {
	    in.ignore( "EXTRUSION", tok.c_str() );
	}
    }

    return stream;
}
