#line 356 "nws/halfspace.nw"
#include <string>
#include <sstream>
#include <math.h>
#include "input.h"
#include "halfspace.h"

const char* halfspace::id = "NKVERSION: halfspace( 2.1.2007.03.21 )"; 

#line 39 "nws/halfspace.nw"
halfspace::halfspace( const scene* ss ) : object( ss )
{
}
#line 51 "nws/halfspace.nw"
halfspace::halfspace( const mathvec<double>& cc ) : object( 0 )
{
    this->center = cc;

    this->orientation.resize( 1 );
    this->orientation[ 0 ] = cc;
    this->orientation[ 0 ].normalize();
}
#line 67 "nws/halfspace.nw"
halfspace::halfspace(
    const mathvec<double>& cc, const mathvec<double>& nn
) : object( 0 )
{
    this->center = cc;

    this->orientation.resize( 1 );
    this->orientation[ 0 ] = nn;
    this->orientation[ 0 ].normalize();
}
#line 87 "nws/halfspace.nw"
halfspace::~halfspace( void ) {
};
#line 99 "nws/halfspace.nw"
void
halfspace::setDimensions( unsigned int dd )
{
    this->object::setDimensions( dd );
}
#line 120 "nws/halfspace.nw"
bool
halfspace::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    
#line 139 "nws/halfspace.nw"
unsigned int len = this->center.size();

mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

mathvec< double > nn( 0.0, len );
if ( rp[ 0 ] > 0.0 ) {
    nn[ 0 ] = 1.0;
} else {
    nn[ 0 ] = -1.0;
}

double tt;
double dd = rd[ 0 ];

if ( ::fabs( dd ) > 0.000001 ) {
    tt = - rp[ 0 ] / dd;
} else {
    tt = - 1.0;
}

#line 128 "nws/halfspace.nw"
    if ( tt > 0.000001 ) {
	
#line 168 "nws/halfspace.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( rp[ 0 ] > 0.0 ) {
    hit.setOutside();
} else {
    hit.setInside();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd.normalize() );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 130 "nws/halfspace.nw"
    } else {
	return false;
    }

    return true;
}
#line 203 "nws/halfspace.nw"
bool
halfspace::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    
#line 139 "nws/halfspace.nw"
unsigned int len = this->center.size();

mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

mathvec< double > nn( 0.0, len );
if ( rp[ 0 ] > 0.0 ) {
    nn[ 0 ] = 1.0;
} else {
    nn[ 0 ] = -1.0;
}

double tt;
double dd = rd[ 0 ];

if ( ::fabs( dd ) > 0.000001 ) {
    tt = - rp[ 0 ] / dd;
} else {
    tt = - 1.0;
}

#line 211 "nws/halfspace.nw"
    hits.clear();

    if ( tt > 0.000001 ) {
	intersection* hh = new intersection();
	intersection& hit( *hh );
	
#line 168 "nws/halfspace.nw"
mathvec< double > spot = rp + ( rd * tt );

if ( rp[ 0 ] > 0.0 ) {
    hit.setOutside();
} else {
    hit.setInside();
}

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd.normalize() );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );
#line 217 "nws/halfspace.nw"
	hits.insert( hits.end(), hh );
    } else if ( rp[ 0 ] <= 0.0 ) {
	intersection* hh = new intersection();
	this->forgeHit( direction, start, *hh );
	hits.insert( hits.end(), hh );
    }

    return hits.size() > 0;
}
#line 237 "nws/halfspace.nw"
double
halfspace::distance(
    const mathvec< double >& start
) const {
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;
    return rp[ 0 ];
}
#line 256 "nws/halfspace.nw"
bool
halfspace::inside( const mathvec< double >& start ) const
{
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
halfspace::_inside( const mathvec< double >& rp ) const
{
    return rp[ 0 ] <= 0.0;
}
#line 282 "nws/halfspace.nw"
std::istream&
operator >> ( std::istream& stream, halfspace& hh )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "base" ) {
	    stream >> (object&)hh >> semicolon;
	} else {
	    in.ignore( "HALFSPACE", tok.c_str() );
	}
    }

    return stream;
}
