#line 282 "nws/image.nw"
#include <string>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "input.h"
#include "image.h"

const char* image::id = "NKVERSION: image( 2.1.2007.03.21 )"; 

#line 216 "nws/image-ppm.nw"
#ifdef USE_PPM
class imagePPM : public image {
public:
    
#line 40 "nws/image-ppm.nw"
imagePPM( void ) : out( 0 ), borderRow( 0 ) {
};
#line 49 "nws/image-ppm.nw"
virtual ~imagePPM( void ) {
    delete this->out;
    delete[] this->borderRow;
};
#line 62 "nws/image-ppm.nw"
virtual void setSize(
    unsigned int ww, unsigned int hh,
    unsigned int dd,
    unsigned int cc
) {
    if ( cc > 3 ) {
	cc = 3;
    } else if ( cc < 3 ) {
	cc = 1;
    }

    this->image::setSize( ww, hh, dd, cc );

    unsigned int len = this->width * this->channelCount;

    delete[] this->borderRow;
    this->borderRow = new unsigned char[ len ];

    for ( unsigned int ii=0; ii < len; ++ii ) {
	this->borderRow[ ii ] = (unsigned char)0xFF;
    }

    this->pixelCount = 0;
};
#line 97 "nws/image-ppm.nw"
virtual void add( const mathvec< double >& pixel ) {
    this->checkOutput();
    if ( pixel.size() <= this->channelCount ) {
	unsigned char lastVal = 0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    int val = (int)( pixel[ ii ] * 255.0 + 0.5 );
	    lastVal = (unsigned char)( val<0 ) ? 0 : ( val>255 ) ? 255 : val ;
	    this->out->put( lastVal );
	}
	for ( unsigned int ii=pixel.size(); ii < this->channelCount; ++ii ) {
	    this->out->put( lastVal );
	}
    } else {
	int total = 0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    int val = (int)( pixel[ ii ] * 255.0 + 0.5 );
	    total += ( val < 0 ) ? 0 : ( val > 255 ) ? 255 : val ;
	}
	unsigned char val = (unsigned char)( total / pixel.size() );
	for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
	    this->out->put( val );
	}
    }
};
#line 126 "nws/image-ppm.nw"
virtual void addBorderPixel( void ) {
    this->checkOutput();
    this->out->write( (const char*)borderRow, this->channelCount );
};
#line 135 "nws/image-ppm.nw"
virtual void addBorderRow( void ) {
    this->checkOutput( this->width );
    this->out->write( (const char*)borderRow, this->width*this->channelCount );
};
#line 183 "nws/image-ppm.nw"
virtual std::istream& get( std::istream& stream )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "common" ) {
	    this->image::get( stream );
	    stream >> semicolon;
	} else {
	    in.ignore( "PPM", tok.c_str() );
	}
    }

    return stream;
};
#line 220 "nws/image-ppm.nw"
protected:
    
#line 144 "nws/image-ppm.nw"
void checkOutput( unsigned int count = 1 ) {
    if ( this->pixelCount < count ) {
	delete this->out;
	this->out = 0;
	this->pixelCount = this->width * this->height - count;
    } else {
	this->pixelCount -= count;
    }
    if ( this->out == 0 ) {
	std::string fname = this->getNumberedBasename() + ".ppm";
	this->out = new std::ofstream( fname.c_str() );
	if ( this->channelCount == 1 ) {
	    *this->out << "P5" << std::endl;
	} else {
	    *this->out << "P6" << std::endl;
	}
	*this->out << this->width << " " << this->height << std::endl;
	*this->out << "255" << std::endl;
    }
};
#line 222 "nws/image-ppm.nw"
    
#line 20 "nws/image-ppm.nw"
std::ostream* out;
#line 26 "nws/image-ppm.nw"
unsigned char* borderRow;
#line 32 "nws/image-ppm.nw"
unsigned int pixelCount;
#line 223 "nws/image-ppm.nw"
public:
    static const char* id;
};

const char* imagePPM::id = "NKVERSION: imagePPM( 2.1.2007.03.21 )"; 
#endif
#line 301 "nws/image-png.nw"
#ifdef USE_PNG
#include <png.h>

class imagePNG : public image {
public:
    
#line 51 "nws/image-png.nw"
imagePNG( void ) : out( 0 ), borderRow( 0 ), tmpBuffer( 0 ), pixelBuffer( 0 ),
    pngPtr( 0 ), infoPtr( 0 )
{
};
#line 62 "nws/image-png.nw"
virtual ~imagePNG( void ) {
    if ( this->pngPtr != 0 ) {
	::png_write_end( this->pngPtr, this->infoPtr );
	::png_destroy_write_struct( &this->pngPtr, &this->infoPtr );
    }
    delete this->out;
    delete[] this->pixelBuffer;
    delete[] this->borderRow;
    delete[] this->tmpBuffer;
};
#line 81 "nws/image-png.nw"
virtual void setSize(
    unsigned int ww, unsigned int hh,
    unsigned int dd,
    unsigned int cc
) {
    if ( cc > 4 ) {
	cc = 4;
    }

    this->image::setSize( ww, hh, dd, cc );

    unsigned int len = this->width * this->channelCount;

    delete[] this->borderRow;
    this->borderRow = new unsigned char[ len ];

    delete[] this->tmpBuffer;
    this->tmpBuffer = new unsigned char[ len ];

    delete[] this->pixelBuffer;
    this->pixelBuffer = new png_byte[ len ];

    for ( unsigned int ii=0; ii < len; ++ii ) {
	this->borderRow[ ii ] = (unsigned char)0xFF;
    }

    this->pixelCount = 0;
};
#line 120 "nws/image-png.nw"
virtual void add( const mathvec< double >& pixel ) {
    unsigned int curSample = 0;
    if ( pixel.size() <= this->channelCount ) {
	unsigned char lastVal = 0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    int val = (int)( pixel[ ii ] * 255.0 + 0.5 );
	    lastVal = (unsigned char)( val<0 ) ? 0 : ( val>255 ) ? 255 : val ;
	    this->tmpBuffer[ curSample++ ] = lastVal;
	}
	for ( unsigned int ii=pixel.size(); ii < this->channelCount; ++ii ) {
	    this->tmpBuffer[ curSample++ ] = lastVal;
	}
    } else {
	int total = 0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    int val = (int)( pixel[ ii ] * 255.0 + 0.5 );
	    total += ( val < 0 ) ? 0 : ( val > 255 ) ? 255 : val ;
	}
	unsigned char val = (unsigned char)( total / pixel.size() );
	for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
	    this->tmpBuffer[ curSample++ ] = val;
	}
    }
    this->write( this->tmpBuffer, 1 );
};
#line 150 "nws/image-png.nw"
virtual void addBorderPixel( void ) {
    this->write( this->borderRow, 1 );
};
#line 158 "nws/image-png.nw"
virtual void addBorderRow( void ) {
    this->write( this->borderRow, this->width );
};
#line 268 "nws/image-png.nw"
virtual std::istream& get( std::istream& stream )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "common" ) {
	    this->image::get( stream );
	    stream >> semicolon;
	} else {
	    in.ignore( "PNG", tok.c_str() );
	}
    }

    return stream;
};
#line 307 "nws/image-png.nw"
protected:
    
#line 166 "nws/image-png.nw"
void write( const unsigned char* buffer, unsigned int count ) {
    if ( this->pixelCount < count ) {
	if ( this->pngPtr != 0 ) {
	    ::png_write_end( this->pngPtr, this->infoPtr );
	    ::png_destroy_write_struct( &this->pngPtr, &this->infoPtr );
	}
	delete this->out;
	this->out = 0;
	this->pixelCount = this->width * this->height - count;
    } else {
	this->pixelCount -= count;
    }
    if ( this->out == 0 ) {
	std::string fname = this->getNumberedBasename() + ".png";
	this->out = new std::ofstream( fname.c_str() );

	this->pngPtr = ::png_create_write_struct(
	    PNG_LIBPNG_VER_STRING, 0, 0, 0
	);
	if ( this->pngPtr == 0 ) {
	    std::cerr << "Unable to create png write structure" << std::endl;
	    return;
	}

	this->infoPtr = ::png_create_info_struct( this->pngPtr );
	if ( this->infoPtr == 0 ) {
	    std::cerr << "Unable to create png info structure" << std::endl;
	    ::png_destroy_write_struct( &this->pngPtr, 0 );
	    return;
	}

	if ( ::setjmp( this->pngPtr->jmpbuf ) ) {
	    std::cerr << "Something wrong during write png" << std::endl;
	    ::png_destroy_write_struct( &this->pngPtr, &this->infoPtr );
	}

	::png_set_write_fn(
	    this->pngPtr, this->out, imagePNG::writeData, imagePNG::flushData
	);
	::png_set_compression_level( this->pngPtr, Z_BEST_COMPRESSION );

	static int colorTypes[] = {
	    0,
	    PNG_COLOR_TYPE_GRAY, PNG_COLOR_TYPE_GRAY_ALPHA,
	    PNG_COLOR_TYPE_RGB, PNG_COLOR_TYPE_RGB_ALPHA
	};

	::png_set_IHDR(
	    this->pngPtr, this->infoPtr, this->width, this->height, 8,
	    colorTypes[ this->channelCount ], PNG_INTERLACE_NONE,
	    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT
	);

	::png_write_info( this->pngPtr, this->infoPtr );

	this->rowSpot = 0;
    }

    count *= this->channelCount;
    for ( unsigned int ii=0; ii < count; ++ii ) {
	this->pixelBuffer[ rowSpot++ ] = buffer[ ii ];
    }

    if ( rowSpot >= this->width * this->channelCount ) {
	::png_write_row( this->pngPtr, this->pixelBuffer );
	this->rowSpot = 0;
    }
};
#line 239 "nws/image-png.nw"
static void writeData( png_structp ctx, png_bytep buffer, png_size_t size )
{
    std::ofstream* dst = (std::ofstream*)::png_get_io_ptr( ctx );
    dst->write( (const char*)buffer, size );
}
static void flushData( png_structp ctx )
{
    std::ofstream* dst = (std::ofstream*)::png_get_io_ptr( ctx );
    dst->flush();
}
#line 309 "nws/image-png.nw"
    
#line 20 "nws/image-png.nw"
std::ostream* out;
#line 28 "nws/image-png.nw"
unsigned char* borderRow;
unsigned char* tmpBuffer;
png_bytep pixelBuffer;
unsigned int rowSpot;
#line 36 "nws/image-png.nw"
png_structp pngPtr;
png_infop infoPtr;
#line 43 "nws/image-png.nw"
unsigned int pixelCount;
#line 310 "nws/image-png.nw"
public:
    static const char* id;
};

const char* imagePNG::id = "NKVERSION: imagePNG( 2.1.2007.03.21 )"; 
#endif
#line 292 "nws/image-accum.nw"
class imageAccum : public image {
protected:
    
#line 34 "nws/image-accum.nw"
typedef enum {
    MINIMUM, MAXIMUM, AVERAGE, FIRST
} MethodType;
#line 295 "nws/image-accum.nw"
public:
    
#line 57 "nws/image-accum.nw"
imageAccum( void ) : sub( 0 ), accumulator( 0 ) {
};
#line 66 "nws/image-accum.nw"
virtual ~imageAccum( void ) {
    delete this->sub;
    delete[] this->accumulator;
};
#line 79 "nws/image-accum.nw"
virtual void setSize(
    unsigned int ww, unsigned int hh,
    unsigned int dd,
    unsigned int cc
) {
    this->image::setSize( ww, hh, dd, cc );
    this->sub->setSize( ww, hh, 1, cc );
    this->pixelsRemaining = ww * hh * dd;

    unsigned int len = ww * hh * cc;

    this->accumulator = new double[ len ];
    this->curIndex = 0;

    double initialValue = 0.0;
    if ( this->method == MINIMUM ) {
	initialValue = 1.0;
    }

    for ( unsigned int ii=0; ii < len; ++ii ) {
	this->accumulator[ ii ] = initialValue;
    }

    this->curPixel.resize( cc, 0.0 );
    this->allOnes.resize( cc, 1.0 );
};
#line 116 "nws/image-accum.nw"
virtual void add( const mathvec< double >& pixel ) {
    if ( pixel.size() <= this->channelCount ) {
	double lastVal = 0.0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    lastVal = pixel[ ii ];
	    this->curPixel[ ii ] = lastVal;
	}
	for ( unsigned int ii=pixel.size(); ii < this->channelCount; ++ii ) {
	    this->curPixel[ ii ] = lastVal;
	}
    } else {
	double total = 0.0;
	for ( unsigned int ii=0; ii < pixel.size(); ++ii ) {
	    total += pixel[ ii ];
	}
	double val = ( total / (double)pixel.size() );
	for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
	    this->curPixel[ ii ] = val;
	}
    }
    this->add();
};
#line 143 "nws/image-accum.nw"
virtual void addBorderPixel( void ) {
    this->curPixel = this->allOnes;
    this->add();
};
#line 152 "nws/image-accum.nw"
virtual void addBorderRow( void ) {
    this->curPixel = this->allOnes;
    for ( unsigned int ii=0; ii < this->width; ++ii ) {
	this->add();
    }
};
#line 244 "nws/image-accum.nw"
virtual std::istream& get( std::istream& stream )
{
    input in( stream );

    this->method = AVERAGE;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "method" ) {
	    if ( in.get( token, sizeof(token) ) == 0 ) {
		return stream;
	    }
	    std::string meth( token );
	    if ( meth == "minimum" || meth == "min" ) {
		this->method = MINIMUM;
	    } else if ( meth == "maximum" || meth == "max" ) {
		this->method = MAXIMUM;
	    } else if ( meth == "average" || meth == "avg" ) {
		this->method = AVERAGE;
	    } else if ( meth == "first" ) {
		this->method = FIRST;
	    }
	} else if ( tok == "image" ) {
	    this->sub = this->image::read( stream );
	    stream >> semicolon;
	} else {
	    in.ignore( "ACCUM", tok.c_str() );
	}
    }

    return stream;
};
#line 297 "nws/image-accum.nw"
protected:
    
#line 162 "nws/image-accum.nw"
virtual void add( void ) {
    double* spot = &this->accumulator[ this->curIndex ];
    this->curIndex += this->channelCount;

    switch ( this->method ) {
    case MINIMUM: {
	double curTotal = ( this->curPixel ^ this->allOnes );
	double spotTotal = 0.0;
	
#line 219 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spotTotal += spot[ ii ];
}

#line 172 "nws/image-accum.nw"
	if ( curTotal < spotTotal ) {
	    
#line 224 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spot[ ii ] = this->curPixel[ ii ];
}
#line 174 "nws/image-accum.nw"
	}
    } break;
    case MAXIMUM: {
	double curTotal = ( this->curPixel ^ this->allOnes );
	double spotTotal = 0.0;
	
#line 219 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spotTotal += spot[ ii ];
}

#line 181 "nws/image-accum.nw"
	if ( curTotal > spotTotal ) {
	    
#line 224 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spot[ ii ] = this->curPixel[ ii ];
}
#line 183 "nws/image-accum.nw"
	}
    } break;
    case AVERAGE: {
	for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
	    spot[ ii ] += this->curPixel[ ii ] / (double)this->depth;
	}
    } break;
    case FIRST: {
	double spotTotal = 0.0;
	
#line 219 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spotTotal += spot[ ii ];
}

#line 194 "nws/image-accum.nw"
	if ( spotTotal < 0.000001 ) {
	    
#line 224 "nws/image-accum.nw"
for ( unsigned int ii=0; ii < this->channelCount; ++ii ) {
    spot[ ii ] = this->curPixel[ ii ];
}
#line 196 "nws/image-accum.nw"
	}
    } break;
    };

    unsigned int len = this->width * this->height * this->channelCount;

    if ( this->curIndex >= len ) {
	this->curIndex = 0;
    }

    if ( --this->pixelsRemaining == 0 ) {
	mathvec< double > pixel( this->channelCount );
	for ( unsigned int ii=0; ii < len; /*done*/ ) {
	    for ( unsigned int kk=0; kk < this->channelCount; ++kk ) {
		pixel[ kk ] = this->accumulator[ ii++ ];
	    }
	    this->sub->add( pixel );
	}
    }
};
#line 299 "nws/image-accum.nw"
    
#line 43 "nws/image-accum.nw"
image* sub;
double* accumulator;
unsigned int curIndex;
unsigned int pixelsRemaining;
MethodType method;
mathvec< double > curPixel;
mathvec< double > allOnes;
#line 300 "nws/image-accum.nw"
public:
    static const char* id;
};

const char* imageAccum::id = "NKVERSION: imageAccum( 2.1.2007.03.21 )"; 
#line 71 "nws/image.nw"
image::image( void ) : currentImage( 0 )
{
}
#line 86 "nws/image.nw"
image::~image( void )
{
}
#line 108 "nws/image.nw"
void
image::setSize(
    unsigned int ww, unsigned int hh,
    unsigned int dd,
    unsigned int cc
)
{
    this->width = ww;
    this->height = hh;
    this->depth = dd;
    this->channelCount = cc;
}
#line 123 "nws/image.nw"
void
image::setBasename( const std::string& bb )
{
    this->basename = bb;
}
#line 141 "nws/image.nw"
std::string
image::getNumberedBasename( void )
{
    std::ostringstream fname;
    fname << this->basename;
    if ( this->depth > 1 || this->currentImage > 1 ) {
	unsigned int digits
	    = (unsigned int)( ::log10( (double)this->depth ) + 1.0000001 );
	fname << std::setw( digits ) << std::setfill( '0' );
	fname << this->currentImage++;
    }
    return fname.str();
}
#line 176 "nws/image.nw"
image*
image::read( std::istream& stream )
{
    input in( stream );

    if ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return 0;
	}

	std::string tok( token );

	
#line 171 "nws/image-ppm.nw"
#ifdef USE_PPM
    if ( tok == "ppm" ) {
	image* nn = new imagePPM();
	nn->get( stream );
	return nn;
    } else
#endif
#line 256 "nws/image-png.nw"
#ifdef USE_PNG
    if ( tok == "png" ) {
	image* nn = new imagePNG();
	nn->get( stream );
	return nn;
    } else
#endif
#line 234 "nws/image-accum.nw"
if ( tok == "accumulator" || tok == "accum" ) {
    image* nn = new imageAccum();
    nn->get( stream );
    return nn;
} else
#line 191 "nws/image.nw"
	{
	    in.ignore( "IMAGE FACTORY", tok.c_str() );
	}
    }

    return 0;
}
#line 207 "nws/image.nw"
std::istream&
image::get( std::istream& stream )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );

	if ( tok == "basename" || tok == "filename" ) {
	    in.get( token, sizeof(token) );
	    this->basename = token;
	} else {
	    in.ignore( "IMAGE", tok.c_str() );
	}
    }

    return stream;
}
