#line 463 "nws/csg-inter.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "inter.h"

const char* csginter::id = "NKVERSION: csginter( 2.1.2007.03.21 )"; 

#line 44 "nws/csg-inter.nw"
csginter::csginter( const scene* ss ) : object( ss )
{
}
#line 57 "nws/csg-inter.nw"
csginter::~csginter( void ) {
    std::vector< object* >::iterator it;
    for ( it=this->objs.begin(); it != this->objs.end(); ++it ) {
	delete *it;
    }
};
#line 74 "nws/csg-inter.nw"
void
csginter::setDimensions( unsigned int dd )
{
    std::vector< object* >::iterator it;
    for ( it=this->objs.begin(); it != this->objs.end(); ++it ) {
	(*it)->setDimensions( dd );
    }
    this->object::setDimensions( dd );
}
#line 99 "nws/csg-inter.nw"
bool
csginter::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    std::vector< intersection* > hits;
    bool ret = this->intersect( direction, start, hits );
    if ( ret ) {
	hit = *hits[ 0 ];
	std::vector< intersection* >::iterator it;
	for ( it = hits.begin(); it != hits.end(); ++it ) {
	    delete *it;
	}
    }
    return ret;
}
#line 135 "nws/csg-inter.nw"
bool
csginter::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    hits.clear();
    
#line 153 "nws/csg-inter.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;
#line 143 "nws/csg-inter.nw"
    
#line 164 "nws/csg-inter.nw"
unsigned int len = this->objs.size();

std::vector< std::vector< intersection* > > hitLists( len );
std::vector< unsigned int > hitStarts( len );

bool allHit = true;
bool amOutside = false;

for ( unsigned int ii=0; allHit && ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    object* obj = this->objs[ ii ];

    if ( obj->intersect( rd, rp, curHits ) ) {
	if ( ! amOutside ) {
	    amOutside = curHits[ 0 ]->isOutside();
	}
    } else {
	allHit = false;
    }
    hitStarts[ ii ] = 0;
}
#line 144 "nws/csg-inter.nw"
    if ( allHit ) {
	
#line 188 "nws/csg-inter.nw"
intersection* bestHit;
unsigned int bestII = 0;
unsigned int bestJJ = 0;
do {
    bestHit = 0;
    if ( amOutside ) {
	
#line 207 "nws/csg-inter.nw"
bool done = false;
while ( !done ) {
    
#line 215 "nws/csg-inter.nw"
bool notSet = true;
double dist = MAXFLOAT;
for ( unsigned int ii=0; ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    unsigned int hStart = hitStarts[ ii ];
    unsigned int hLen = curHits.size();
    if ( hStart < hLen ) {
	intersection* hit = curHits[ hStart ];
	if ( hit->isOutside() ) {
	    double curDist = hit->getDistance();
	    if ( notSet || curDist < dist ) {
		bestII = ii;
		bestJJ = hStart;
		bestHit = hit;
		dist = curDist - 0.000001;
		notSet = false;
	    }
	}
    }
}
if ( bestHit == 0 ) {
    break;
} else {
    hitStarts[ bestII ] = bestJJ + 1;
}
#line 210 "nws/csg-inter.nw"
    
#line 244 "nws/csg-inter.nw"
bool insideAll = true;
for ( unsigned int ii=0; insideAll && ii < len; ++ii ) {
    if ( ii == bestII ) {
	continue;
    }
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    unsigned int hLen = curHits.size();

    for ( unsigned int jj=hitStarts[ ii ]; jj < hLen; ++jj ) {
	intersection* hit = curHits[ jj ];
	double curDist = hit->getDistance();
	if ( curDist <= dist ) {
	    ++hitStarts[ ii ];
	} else if ( hit->isOutside() ) {
	    insideAll = false;
	    break;
	} else {
	    break;
	}
    }
    if ( hitStarts[ ii ] == hLen && hLen > 0 && curHits[ hLen-1  ] != 0
    && curHits[ hLen-1  ]->isInside() ) {
	insideAll = false;
    }
}
if ( !insideAll ) {
    bestHit = 0;
} else {
    done = true;
}
#line 211 "nws/csg-inter.nw"
}
#line 195 "nws/csg-inter.nw"
    } else {
	
#line 277 "nws/csg-inter.nw"
bool notSet = true;
double dist = MAXFLOAT;
for ( unsigned int ii=0; ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    unsigned int hStart = hitStarts[ ii ];
    unsigned int hLen = curHits.size();
    if ( hStart < hLen ) {
	intersection* hit = curHits[ hStart ];
	if ( hit->isInside() ) {
	    double curDist = hit->getDistance();
	    if ( notSet || curDist < dist ) {
		bestII = ii;
		bestJJ = hStart;
		bestHit = hit;
		dist = curDist - 0.000001;
		notSet = false;
	    }
	}
    }
}
#line 197 "nws/csg-inter.nw"
	
#line 300 "nws/csg-inter.nw"
if ( bestHit != 0 ) {
    for ( unsigned int ii=0; ii < len; ++ii ) {
	if ( ii == bestII ) {
	    continue;
	}
	std::vector< intersection* >& curHits( hitLists[ ii ] );
	unsigned int hLen = curHits.size();

	for ( unsigned int jj=hitStarts[ ii ]; jj < hLen; ++jj ) {
	    intersection* hit = curHits[ jj ];
	    double curDist = hit->getDistance();
	    if ( curDist <= dist ) {
		++hitStarts[ ii ];
	    }
	}
    }
}
#line 198 "nws/csg-inter.nw"
    }
    if ( bestHit != 0 ) {
	
#line 320 "nws/csg-inter.nw"
std::vector< intersection* >& curHits( hitLists[ bestII ] );
curHits[ bestJJ ] = 0;
hitStarts[ bestII ] = bestJJ+1;
bestHit->reorient( this->scale, this->orientation, this->center );
bestHit->setDistance( ( bestHit->getPosition() - start ).magnitude() );
if ( this->gotColor ) {
    bestHit->recolor( this->chroma );
}
hits.insert( hits.end(), bestHit );
#line 201 "nws/csg-inter.nw"
    }
    amOutside = !amOutside;
} while ( bestHit != 0 );
#line 146 "nws/csg-inter.nw"
    }
    
#line 332 "nws/csg-inter.nw"
for ( unsigned int ii=0; ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    unsigned int hLen = curHits.size();
    for ( unsigned int jj=0; jj < hLen; ++jj ) {
	delete curHits[ jj ];
    }
    curHits.clear();
}
#line 148 "nws/csg-inter.nw"
    return hits.size() > 0;
}
#line 350 "nws/csg-inter.nw"
bool
csginter::inside( const mathvec< double >& start ) const
{
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
csginter::_inside( const mathvec< double >& rp ) const
{
    unsigned int len = this->objs.size();
    bool in = len > 0;

    for ( unsigned int ii=0; in && ii < len; ++ii ) {
	in = this->objs[ ii ]->inside( rp );
    }

    return in;
}
#line 382 "nws/csg-inter.nw"
std::istream&
operator >> ( std::istream& stream, csginter& cc )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "object" ) {
	    object* nn = object::read( stream, cc.myScene );
	    if ( nn != 0 ) {
		cc.objs.insert( cc.objs.end(), nn );
	    } else {
		std::cerr << "INTER failed to get an object!" << std::endl;
	    }
	} else if ( tok == "base" ) {
	    stream >> (object&)cc >> semicolon;
	} else {
	    in.ignore( "INTER", tok.c_str() );
	}
    }

    return stream;
}
