#line 278 "nws/light.nw"
#include <string>
#include <math.h>
#include "input.h"
#include "light.h"

const char* light::id = "NKVERSION: light( 2.1.2007.03.21 )"; 

#line 63 "nws/light.nw"
light::light( const scene* ss )
    : myScene( ss ), chroma( ss ),
	falloff( 0.0 ), outerCosine(-2.0), cosineRange(-1.0)
{
}
#line 81 "nws/light.nw"
void
light::setDimensions( unsigned int dd )
{
    this->position.resize( dd, 0.0 );
    this->chroma.setDimensions( dd );
    this->direction.resize( dd, 0.0 );

    double dmag = this->direction.magnitude();
    if ( dmag >= 0.000001 ) {
	this->direction /= dmag;
    }
}
#line 134 "nws/light.nw"
void
light::evaluate(
    const mathvec< double >& pos,
    const mathvec< double >& norm,
    const mathvec< double >& dir,
    double contribution, unsigned int height,
    mathvec< double >& pixel
) const {
    this->chroma.evaluate(
	pos, norm, dir, contribution, height, pixel
    );

    if ( this->outerCosine > -1.5 ) {
	double cc = - ( dir ^ this->direction );
	double diff = cc - this->outerCosine;
	if ( diff < 0.000001 ) {
	    pixel *= 0.0;
	} else if ( diff < this->cosineRange ) {
	    pixel *= diff / this->cosineRange;
	}
    }
}
#line 173 "nws/light.nw"
std::istream&
operator >> ( std::istream& stream, light& ll )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "position" || tok == "center" ) {
	    stream >> ll.position >> semicolon;
	} else if ( tok == "direction" ) {
	    stream >> ll.direction >> semicolon;
	} else if ( tok == "falloff" ) {
	    stream >> ll.falloff >> semicolon;
	} else if ( tok == "angles" ) {
	    
#line 211 "nws/light.nw"
double inner;
char comma;
double outer;
stream >> inner >> comma >> outer >> semicolon;

inner = ::cos( inner * M_PI / 180.0 );
outer = ::cos( outer * M_PI / 180.0 );

if ( inner < outer ) {
    double tmp = inner;
    inner = outer;
    outer = tmp;
}

ll.outerCosine = outer;
ll.cosineRange = inner - outer;
#line 196 "nws/light.nw"
	} else if ( tok == "color" ) {
	    stream >> ll.chroma >> semicolon;
	} else {
	    in.ignore( "LIGHT", tok.c_str() );
	}
    }

    return stream;
}
