#line 869 "nws/mathvec.nw"
#ifndef _NKLEIN_MATHVEC_H_
#define _NKLEIN_MATHVEC_H_ 1

#include <iostream>
#include <vector>
#include <cmath>
#include "basevec.h"

    
#line 842 "nws/mathvec.nw"
template < typename TT > class mathvec;
#line 878 "nws/mathvec.nw"
    
#line 246 "nws/mathvec.nw"
template < typename TT > const mathvec< TT > operator * (
    const TT& alpha, const mathvec< TT >& vv
);

#line 880 "nws/mathvec.nw"
    
#line 845 "nws/mathvec.nw"
template < typename TT >
class mathvec : public basevec< TT > {
public:
    
#line 90 "nws/mathvec.nw"
mathvec();
#line 101 "nws/mathvec.nw"
mathvec( unsigned int len );
#line 114 "nws/mathvec.nw"
mathvec( const TT value, unsigned int len );
#line 126 "nws/mathvec.nw"
mathvec( const TT array[], unsigned int len );
#line 139 "nws/mathvec.nw"
mathvec( const mathvec< TT >& other );
#line 153 "nws/mathvec.nw"
mathvec( const char* str );
#line 172 "nws/mathvec.nw"
bool operator != ( const mathvec< TT >& other ) const;
bool operator == ( const mathvec< TT >& other ) const;
#line 220 "nws/mathvec.nw"
const mathvec< TT >& operator *= ( const TT& alpha );
#line 251 "nws/mathvec.nw"
friend const mathvec< TT > operator * <TT> (
	const TT& alpha, const mathvec< TT >& vv
    );
const mathvec< TT > operator * ( const TT& alpha ) const;
#line 282 "nws/mathvec.nw"
const mathvec< TT >& operator /= ( const TT& other );
#line 311 "nws/mathvec.nw"
const mathvec< TT > operator / ( const TT& alpha );
#line 330 "nws/mathvec.nw"
const mathvec< TT >& operator += ( const mathvec< TT >& other );
#line 352 "nws/mathvec.nw"
const mathvec< TT > operator + ( const mathvec< TT >& other ) const;
#line 371 "nws/mathvec.nw"
const mathvec< TT >& operator -= ( const mathvec< TT >& other );
#line 394 "nws/mathvec.nw"
const mathvec< TT > operator - ( const mathvec< TT >& other ) const;
#line 413 "nws/mathvec.nw"
const mathvec< TT >& operator *= ( const mathvec< TT >& other );
#line 436 "nws/mathvec.nw"
const mathvec< TT > operator * ( const mathvec< TT >& other ) const;
#line 455 "nws/mathvec.nw"
const mathvec< TT >& operator /= ( const mathvec< TT >& other );
#line 484 "nws/mathvec.nw"
const mathvec< TT > operator / ( const mathvec< TT >& other ) const;
#line 504 "nws/mathvec.nw"
const TT operator ^ ( const mathvec< TT >& other ) const;
#line 532 "nws/mathvec.nw"
TT magnitude( void ) const;
mathvec< TT >& normalize( void );
#line 687 "nws/mathvec.nw"
mathvec< TT >& rotate( const basevec< mathvec< TT > >& matrix );
#line 713 "nws/mathvec.nw"
mathvec< TT >& antirotate( const basevec< mathvec< TT > >& matrix );
#line 743 "nws/mathvec.nw"
mathvec< TT > solve( const basevec< mathvec< TT > >& _AA );
#line 849 "nws/mathvec.nw"
public:
    static const char* id;
};
#line 881 "nws/mathvec.nw"
    
#line 93 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec() : basevec< TT >()
{
}
#line 104 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec( unsigned int len ) : basevec< TT >()
{
    TT zero( 0U );
    this->resize( len, zero );
}
#line 117 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec( const TT value, unsigned int len )
    : basevec< TT >( value, len )
{
}
#line 129 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec( const TT array[], unsigned int len )
    : basevec< TT >( array, len )
{
}
#line 142 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec( const mathvec< TT >& other )
    : basevec< TT >( other )
{
}
#line 156 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >::mathvec( const char* str )
    : basevec< TT >( str )
{
}
#line 181 "nws/mathvec.nw"
template < typename TT >
bool
mathvec< TT >::operator != ( const mathvec< TT >& other ) const
{
    if ( this->size() != other.size() ) {
	return true;
    }

    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	if ( (*this)[ ii ] != other[ ii ] ) {
	    return true;
	}
    }

    return false;
}
#line 203 "nws/mathvec.nw"
template < typename TT >
bool
mathvec< TT >::operator == ( const mathvec< TT >& other ) const
{
    return ! ( (*this) != other );
}
#line 226 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator *= ( const TT& alpha )
{
    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] *= alpha;
    }

    return *this;
}
#line 257 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
operator * ( const TT& alpha, const mathvec< TT >& vv )
{
    mathvec< TT > result( vv );
    result *= alpha;

    return result;
}
#line 268 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator * ( const TT& alpha ) const
{
    mathvec< TT > result( *this );
    result *= alpha;

    return result;
}
#line 288 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator /= ( const TT& alpha )
{
    unsigned int len( this->size() );

    TT zero( 0U );
    if ( alpha != zero ) {
	for ( unsigned int ii=0; ii < len; ++ii ) {
	    (*this)[ ii ] /= alpha;
	}
    } else {
	for ( unsigned int ii=0; ii < len; ++ii ) {
	    (*this)[ ii ] = zero;
	}
    }

    return *this;
}
#line 314 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator / ( const TT& alpha )
{
    mathvec< TT > result( *this );
    result /= alpha;

    return result;
}
#line 337 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator += ( const mathvec< TT >& other )
{
    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] += other[ ii ];
    }

    return *this;
}
#line 355 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator + ( const mathvec< TT >& other ) const
{
    mathvec< TT > result( *this );
    result += other;

    return result;
}
#line 379 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator -= ( const mathvec< TT >& other )
{
    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] -= other[ ii ];
    }

    return *this;
}
#line 397 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator - ( const mathvec< TT >& other ) const
{
    mathvec< TT > result( *this );
    result -= other;

    return result;
}
#line 421 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator *= ( const mathvec< TT >& other )
{
    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	(*this)[ ii ] *= other[ ii ];
    }

    return *this;
}
#line 439 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator * ( const mathvec< TT >& other ) const
{
    mathvec< TT > result( *this );
    result *= other;

    return result;
}
#line 464 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >&
mathvec< TT >::operator /= ( const mathvec< TT >& other )
{
    TT zero( 0U );
    unsigned int len( this->size() );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	if ( other[ ii ] != zero ) {
	    (*this)[ ii ] /= other[ ii ];
	} else {
	    (*this)[ ii ] = zero;
	}
    }

    return *this;
}
#line 487 "nws/mathvec.nw"
template < typename TT >
const mathvec< TT >
mathvec< TT >::operator / ( const mathvec< TT >& other ) const
{
    mathvec< TT > result( *this );
    result /= other;

    return result;
}
#line 510 "nws/mathvec.nw"
template < typename TT >
const TT
mathvec< TT >::operator ^ ( const mathvec< TT >& other ) const
{
    unsigned int len( this->size() );
    TT result( 0U );

    for ( unsigned int ii=0; ii < len; ++ii ) {
	result += (*this)[ ii ] * other[ ii ];
    }

    return result;
}
#line 539 "nws/mathvec.nw"
template < typename TT >
TT
mathvec< TT >::magnitude( void ) const
{
    return (TT)std::sqrt( (double)( (*this) ^ (*this) ) );
}
#line 551 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >&
mathvec< TT >::normalize( void )
{
    *this /= this->magnitude();
    return *this;
}
#line 690 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >&
mathvec< TT >::rotate( const basevec< mathvec< TT > >& matrix )
{
    mathvec< TT >& res( *this );
    mathvec< TT > orig( *this );

    for ( unsigned int ii=0; ii < this->size(); ++ii ) {
	TT value = 0;
	for ( unsigned int jj=0; jj < this->size(); ++jj ) {
	    value += orig[ jj ] * matrix[ jj ][ ii ];
	}
	res[ ii ] = value;
    }


    return res;
}
#line 716 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >&
mathvec< TT >::antirotate( const basevec< mathvec< TT > >& matrix )
{
    mathvec< TT >& res( *this );
    mathvec< TT > orig( *this );

    for ( unsigned int ii=0; ii < this->size(); ++ii ) {
	const mathvec< TT >& col( matrix[ ii ] );
	TT value = 0;
	for ( unsigned int jj=0; jj < this->size(); ++jj ) {
	    value += orig[ jj ] * col[ jj ];
	}
	res[ ii ] = value;
    }

    return res;
}
#line 746 "nws/mathvec.nw"
template < typename TT >
mathvec< TT >
mathvec< TT >::solve( const basevec< mathvec< TT > >& _AA )
{
    unsigned int nn = this->size();
    mathvec< TT > xx( nn );

    
#line 767 "nws/mathvec.nw"
mathvec< unsigned int > perm( nn );
for ( unsigned int ii=0; ii < nn; ++ii ) {
    perm[ ii ] = ii;
}

basevec< mathvec< TT > > AA( _AA );
mathvec< TT > bb( *this );

static const TT zero = (TT)0;
static const TT minusOne = (TT)-1;

#line 755 "nws/mathvec.nw"
    for ( unsigned int jj=0; jj < nn-1; ++jj ) {
	
#line 780 "nws/mathvec.nw"
TT pivot = AA[ perm[ jj ] ][ jj ];
if ( pivot < zero ) {
    pivot *= minusOne;
}
unsigned int piv = jj;
unsigned int ptmp = perm[ jj ];
for ( unsigned int ii=jj+1; ii < nn; ++ii ) {
    TT cur = AA[ perm[ ii ] ][ jj ];
    if ( cur < zero ) {
	cur *= minusOne;
    }
    if ( cur > pivot ) {
	pivot = cur;
	piv = ii;
    }
}

perm[ jj ] = perm[ piv ];
perm[ piv ] = ptmp;

if ( (double)pivot < 0.000001 ) {
    return xx;
}
#line 757 "nws/mathvec.nw"
	
#line 806 "nws/mathvec.nw"
unsigned int pj = perm[ jj ];
const mathvec< TT >& baseRow( AA[ pj ] );
for ( unsigned int ii=jj+1; ii < nn; ++ii ) {
    unsigned int pi = perm[ ii ];
    mathvec< TT >& curRow( AA[ pi ] );
    TT mm = curRow[ jj ] / baseRow[ jj ];
    curRow[ jj ] = zero;
    bb[ pi ] -= mm * bb[ pj ];
    for ( unsigned int kk=jj+1; kk < nn; ++kk ) {
	curRow[ kk ] -= mm * baseRow[ kk ];
    }
}
#line 758 "nws/mathvec.nw"
    }

    
#line 821 "nws/mathvec.nw"
for ( unsigned int kk=0; kk < nn; ++kk ) {
    unsigned int ii = ( nn - 1 ) - kk;
    unsigned int pi = perm[ ii ];
    const mathvec< TT >& curRow( AA[ pi ] );

    xx[ ii ] = bb[ pi ];
    for ( unsigned int jj=ii+1; jj < nn; ++jj ) {
	xx[ ii ] -= curRow[ jj ] * xx[ jj ];
    }
    xx[ ii ] /= curRow[ ii ];
}

#line 762 "nws/mathvec.nw"
    return xx;
}

#line 883 "nws/mathvec.nw"
    
#line 593 "nws/mathvec.nw"
template < typename TT > bool orthogonalize(
    basevec< mathvec< TT > >& matrix,
    unsigned int dims
);
#line 662 "nws/mathvec.nw"
template < typename TT > void transpose(
    basevec< mathvec< TT > >& matrix
);
#line 884 "nws/mathvec.nw"
    
#line 600 "nws/mathvec.nw"
template < typename TT >
bool orthogonalize(
    basevec< mathvec< TT > >& matrix,
    unsigned int dims
)
{
    unsigned int originalSize = matrix.size();
    
#line 614 "nws/mathvec.nw"
for ( unsigned int ii=0; ii < originalSize; ++ii ) {
    matrix[ ii ].resize( dims, 0.0 );
    
#line 625 "nws/mathvec.nw"
mathvec< double > vv( matrix[ ii ] );
for ( unsigned int jj=0; jj < ii; ++jj ) {
    matrix[ ii ] -= ( vv ^ matrix[ jj ] ) * matrix[ jj ];
}
#line 617 "nws/mathvec.nw"
    
#line 632 "nws/mathvec.nw"
double mag = matrix[ ii ].magnitude();
bool zero = ( mag < 0.000001 );
if ( ! zero ) {
    matrix[ ii ] /= mag;
}
#line 618 "nws/mathvec.nw"
    if ( zero ) {
	return false;
    }
}
#line 608 "nws/mathvec.nw"
    
#line 640 "nws/mathvec.nw"
matrix.resize( dims );
unsigned int currentGuessColumn = 0;
for ( unsigned int ii=originalSize; ii < dims; ++ii ) {
    matrix[ ii ].resize( dims, 0.0 );
    
#line 654 "nws/mathvec.nw"
matrix[ ii ] *= 0.0;
matrix[ ii ][ currentGuessColumn++ ] = 1.0;
#line 645 "nws/mathvec.nw"
    
#line 625 "nws/mathvec.nw"
mathvec< double > vv( matrix[ ii ] );
for ( unsigned int jj=0; jj < ii; ++jj ) {
    matrix[ ii ] -= ( vv ^ matrix[ jj ] ) * matrix[ jj ];
}
#line 646 "nws/mathvec.nw"
    
#line 632 "nws/mathvec.nw"
double mag = matrix[ ii ].magnitude();
bool zero = ( mag < 0.000001 );
if ( ! zero ) {
    matrix[ ii ] /= mag;
}
#line 647 "nws/mathvec.nw"
    if ( zero ) {
	--ii;
    }
}
#line 609 "nws/mathvec.nw"
    return true;
}
#line 668 "nws/mathvec.nw"
template < typename TT >
void transpose(
    basevec< mathvec< TT > >& matrix
)
{
    unsigned int originalSize = matrix.size();
    for ( unsigned int ii=0; ii < originalSize-1; ++ii ) {
	for ( unsigned int jj=ii+1; jj < originalSize; ++jj ) {
	    TT tt = matrix[ ii ][ jj ];
	    matrix[ ii ][ jj ] = matrix[ jj ][ ii ];
	    matrix[ jj ][ ii ] = tt;
	}
    }
}

#line 886 "nws/mathvec.nw"
#endif /*_NKLEIN_MATHVEC_H_*/
