#line 280 "nws/nkthread.nw"
#include <iostream>
#include <vector>
#include "nkthread.h"

#line 130 "nws/thread-worker.nw"
extern "C" void* ThreadWorkerLoopStarter( void* arg );
#line 25 "nws/thread-work.nw"
thread::work::work( void ) : done( true )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_init( &this->mutex, 0 );
    ::pthread_cond_init( &this->condition, 0 );
#endif
}
#line 42 "nws/thread-work.nw"
thread::work::~work( void )
{
#ifdef USE_PTHREADS
    ::pthread_cond_destroy( &this->condition );
    ::pthread_mutex_destroy( &this->mutex );
#endif
}
#line 64 "nws/thread-work.nw"
void
thread::work::start( void )
{
    this->done = false;

    thread* mgr = thread::instance();
    mgr->start( this );
}
#line 79 "nws/thread-work.nw"
void
thread::work::wait( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_lock( &this->mutex );
	while ( !this->done ) {
	    ::pthread_cond_wait( &this->condition, &this->mutex );
	}
    ::pthread_mutex_unlock( &this->mutex );
#endif
}
#line 99 "nws/thread-work.nw"
void
thread::work::finish( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_lock( &this->mutex );
#endif
	this->done = true;
#ifdef USE_PTHREADS
	::pthread_cond_broadcast( &this->condition );
    ::pthread_mutex_unlock( &this->mutex );
#endif
}
#line 31 "nws/thread-worker.nw"
thread::worker::worker( void ) : todo( 0 ), done( false )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_init( &this->mutex, 0 );
    ::pthread_cond_init( &this->condition, 0 );
    ::pthread_create( &this->myThread, 0, ThreadWorkerLoopStarter, this );
#endif
}
#line 50 "nws/thread-worker.nw"
thread::worker::~worker( void )
{
    this->lock();
	this->done = true;
	this->wakeup();
    this->unlock();

#ifdef USE_PTHREADS
    ::pthread_join( this->myThread, 0 );
    ::pthread_cond_destroy( &this->condition );
    ::pthread_mutex_destroy( &this->mutex );
#endif
}
#line 82 "nws/thread-worker.nw"
void
thread::worker::wakeup( void )
{
#ifdef USE_PTHREADS
    ::pthread_cond_signal( &this->condition );
#endif
}
#line 97 "nws/thread-worker.nw"
void
thread::worker::lock( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_lock( &this->mutex );
#endif
}
#line 107 "nws/thread-worker.nw"
void
thread::worker::wait( void )
{
#ifdef USE_PTHREADS
    ::pthread_cond_wait( &this->condition, &this->mutex );
#endif
}

#line 116 "nws/thread-worker.nw"
void
thread::worker::unlock( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_unlock( &this->mutex );
#endif
}
#line 134 "nws/thread-worker.nw"
void*
ThreadWorkerLoopStarter( void* arg )
{
    thread::worker* me = (thread::worker*)arg;
    return me->loop();
}
#line 147 "nws/thread-worker.nw"
void*
thread::worker::loop( void )
{
#ifdef USE_PTHREADS
    thread* mgr = thread::instance();

    for (;;) {
	this->todo = 0;
	mgr->makeWorkerAvailable( this );

	this->lock();
	    do {
		if ( this->done ) {
		    this->unlock();
		    return 0;
		}
		this->wait();
	    } while ( this->todo == 0 );
	this->unlock();

	this->todo->perform();
	this->todo->finish();
    }
#endif
    return 0;
}

#line 288 "nws/nkthread.nw"
const char* thread::id = "NKVERSION: thread( 2.1.2007.04.07 )"; 
const char* thread::work::id = "NKVERSION: thread::work( 2.1.2007.04.07 )"; 
const char* thread::worker::id = "NKVERSION: thread::worker( 2.1.2007.04.07 )"; 

#line 40 "nws/nkthread.nw"
thread*
thread::instance( void ) {
    static thread mgr;
    return &mgr;
};
#line 55 "nws/nkthread.nw"
thread::thread( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_init( &this->mutex, 0 );
#endif
}
#line 71 "nws/nkthread.nw"
thread::~thread( void )
{
    this->setWorkerCount( 0 );
#ifdef USE_PTHREADS
    ::pthread_mutex_destroy( &this->mutex );
#endif
}
#line 92 "nws/nkthread.nw"
void
thread::setWorkerCount( unsigned int nn )
{
    std::vector< worker* > tmp;

    this->lock();
	unsigned int original = this->workers.size();
	if ( original < nn ) {
	    for ( unsigned int ii=original; ii < nn; ++ii ) {
		new worker();
	    }
	} else if ( original > nn ) {
	    tmp.resize( original - nn );

	    for ( unsigned int ii=nn; ii < original; ++ii ) {
		tmp[ ii - nn ] = this->workers.front();
		this->workers.pop_front();
	    }
	}

#ifdef USE_PTHREADS
	::pthread_setconcurrency( nn+1 );
#endif
    this->unlock();

    for ( unsigned int ii=0; ii < tmp.size(); ++ii ) {
	delete tmp[ ii ];
    }
}
#line 132 "nws/nkthread.nw"
void
thread::makeWorkerAvailable( worker* ww )
{
    this->lock();
	this->workers.push_back( ww );
    this->unlock();
}
#line 152 "nws/nkthread.nw"
void
thread::start( thread::work* ww )
{
    worker* wr = 0;

    this->lock();
	if ( this->workers.size() > 0 ) {
	    wr = this->workers.front();
	    this->workers.pop_front();
	}
    this->unlock();

    if ( wr == 0 ) {
	ww->perform();
	ww->finish();
    } else {
	wr->lock();
	    wr->setWork( ww );
	    wr->wakeup();
	wr->unlock();
    }
}
#line 186 "nws/nkthread.nw"
void
thread::lock( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_lock( &this->mutex );
#endif
}
#line 196 "nws/nkthread.nw"
void
thread::unlock( void )
{
#ifdef USE_PTHREADS
    ::pthread_mutex_unlock( &this->mutex );
#endif
}
#line 215 "nws/nkthread.nw"
void
thread::wait( void* cond )
{
#ifdef USE_PTHREADS
    ::pthread_cond_wait( (pthread_cond_t*)cond, &this->mutex );
#endif
}
