#line 398 "nws/object.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "object.h"

#line 18 "nws/halfspace.nw"
#include "halfspace.h"
#line 25 "nws/cylinder.nw"
#include "cylinder.h"
#line 25 "nws/quadratic.nw"
#include "quadratic.h"
#line 36 "nws/convex.nw"
#include "inter.h"
#include "convex.h"
#line 34 "nws/coxeter.nw"
#include "inter.h"
#include "convex.h"
#include "coxeter.h"
#line 20 "nws/complement.nw"
#include "complement.h"
#line 20 "nws/csg-union.nw"
#include "union.h"
#line 20 "nws/csg-set.nw"
#include "set.h"
#line 20 "nws/csg-inter.nw"
#include "inter.h"
#line 21 "nws/extrusion.nw"
#include "extrusion.h"

#line 409 "nws/object.nw"
const char* object::id = "NKVERSION: object( 2.1.2007.05.17 )"; 

#line 56 "nws/object.nw"
color object::fakeColor(
    "channels [3]< 0 , 0 , 0 >;\n"
    "reflectiveness 0.0;\n"
    "specularness 0.0;\n"
    "transparency 1.0;\n"
);
#line 72 "nws/object.nw"
object::object( const scene* ss )
    : myScene( ss ), defaultScale( 1.0 ), chroma( ss ), gotColor(false)
{
};
#line 87 "nws/object.nw"
object::~object( void ) {
};
#line 101 "nws/object.nw"
void
object::setDimensions( unsigned int dd )
{
    this->scale.resize( dd, this->defaultScale );
    ::orthogonalize( this->orientation, dd );
    this->center.resize( dd, 0.0 );
    this->chroma.setDimensions( dd );
}
#line 203 "nws/object.nw"
void
object::evaluate(
    const mathvec< double >& position,
    const mathvec< double >& normal,
    const mathvec< double >& direction,
    double contribution, unsigned int height,
    mathvec< double >& pixel
) const {
    this->chroma.evaluate(
	position, normal, direction, contribution, height, pixel
    );
}
#line 234 "nws/object.nw"
void
object::forgeHit(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const
{
    mathvec< double > pos = direction;
    pos *= MAXFLOAT;
    pos += start;

    mathvec< double > norm = direction;
    norm *= -1.0;

    hit.setPosition( pos );
    hit.setNormal( norm );
    hit.setDirection( direction );

    hit.setObjectPosition( pos );
    hit.setObjectNormal( norm );
    hit.setObjectDirection( direction );

    hit.setInside();
    hit.setDistance( MAXFLOAT );
    hit.setObjectColor( &object::fakeColor );
}
#line 270 "nws/object.nw"
std::istream&
operator >> ( std::istream& stream, object& ll )
{
    input in( stream );

    ll.gotColor = false;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "scale" ) {
	    stream >> ll.scale >> semicolon;
	} else if ( tok == "orientation" ) {
	    stream >> ll.orientation >> semicolon;
	} else if ( tok == "antiorientation" ) {
	    stream >> ll.orientation >> semicolon;
	    ::orthogonalize( ll.orientation, ll.orientation[0].size() );
	    ::transpose( ll.orientation );
	} else if ( tok == "center" ) {
	    stream >> ll.center >> semicolon;
	} else if ( tok == "color" ) {
	    stream >> ll.chroma >> semicolon;
	    ll.gotColor = true;
	} else {
	    in.ignore( "OBJECT", tok.c_str() );
	}
    }

    return stream;
}
#line 320 "nws/object.nw"
object*
object::read( std::istream& stream, const scene* ss )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return 0;
	}

	std::string objectType( token );
	char semicolon;

	
#line 21 "nws/halfspace.nw"
if ( objectType == "halfspace" ) {
    halfspace* obj = new halfspace( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 28 "nws/cylinder.nw"
if ( objectType == "cylinder" ) {
    cylinder* obj = new cylinder( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 28 "nws/quadratic.nw"
if ( objectType == "quadratic" ) {
    quadratic* obj = new quadratic( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 40 "nws/convex.nw"
if ( objectType == "convex" ) {
    convex* obj = new convex( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 39 "nws/coxeter.nw"
if ( objectType == "coxeter" ) {
    coxeter* obj = new coxeter( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 23 "nws/complement.nw"
if ( objectType == "complement" ) {
    complement* obj = new complement( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 23 "nws/csg-union.nw"
if ( objectType == "union" ) {
    csgunion* obj = new csgunion( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 23 "nws/csg-set.nw"
if ( objectType == "set" ) {
    csgset* obj = new csgset( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 23 "nws/csg-inter.nw"
if ( objectType == "intersection" || objectType == "inter" ) {
    csginter* obj = new csginter( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 24 "nws/extrusion.nw"
if ( objectType == "extrusion" ) {
    extrusion* obj = new extrusion( ss );
    stream >> *obj >> semicolon;
    return obj;
} else
#line 335 "nws/object.nw"
                               {
	    in.ignore( "OBJECT READER", objectType.c_str() );
	}
    }

    return 0;
}
