#line 306 "nws/portal.nw"
#include <string>
#include <sstream>
#include "input.h"
#include "universe.h"
#include "scene.h"
#include "portal.h"

const char* portal::id = "NKVERSION: portal( 2.1.2007.03.21 )"; 

#line 67 "nws/portal.nw"
portal::portal( const scene* ss ) : myScene( ss ), univ( 0 )
{
}
#line 80 "nws/portal.nw"
portal::~portal( void )
{
}
#line 114 "nws/portal.nw"
void
portal::setDimensions( unsigned int dd )
{
    this->scale.resize( dd, 1.0 );
    ::orthogonalize( this->orientation, dd );
    this->center.resize( dd, 0.0 );
}
#line 140 "nws/portal.nw"
bool
portal::evaluate(
    const mathvec< double >& position,
    const mathvec< double >& direction,
    double contribution, unsigned int height, unsigned int channelCount
) {
    
#line 161 "nws/portal.nw"
if ( this->univ == 0 && this->universeName != "" ) {
    this->univ = this->myScene->getUniverse( this->universeName );
}
if ( this->univ == 0 ) {
    return false;
}
#line 147 "nws/portal.nw"
    
#line 173 "nws/portal.nw"
mathvec< double > rd = direction;
rd.rotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = position;
rp.rotate( this->orientation );
rp /= this->scale;
rp += this->center;
#line 148 "nws/portal.nw"
    
#line 190 "nws/portal.nw"
unsigned int dims = this->univ->getDimensions();
if ( rp.size() != dims ) {
    rp.resize( dims, 0.0 );
    rd.resize( dims, 0.0 );

    double magdir = rd.magnitude();
    if ( magdir >= 0.000001 ) {
	rd /= magdir;
    } else {
	this->pixel *= 0.0;
	return false;
    }
}

#line 150 "nws/portal.nw"
    this->pixel.resize( channelCount, 0.0 );
    return this->univ->evaluate(
	rp, rd, contribution, height-1, this->pixel
    );
}
#line 216 "nws/portal.nw"
std::istream&
operator >> ( std::istream& stream, portal& pp )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "scale" ) {
	    stream >> pp.scale >> semicolon;
	} else if ( tok == "orientation" ) {
	    stream >> pp.orientation >> semicolon;
	} else if ( tok == "center" ) {
	    stream >> pp.center >> semicolon;
	} else if ( tok == "universe" ) {
	    if ( in.get( token, sizeof(token) ) == 0 ) {
		return stream;
	    }
	    pp.universeName = token;
	} else {
	    in.ignore( "PORTAL", tok.c_str() );
	}
    }

    return stream;
}
