#line 414 "nws/quadratic.nw"
#include <string>
#include <sstream>
#include <math.h>
#include "input.h"
#include "quadratic.h"

const char* quadratic::id = "NKVERSION: quadratic( 2.1.2007.03.21 )"; 

#line 49 "nws/quadratic.nw"
quadratic::quadratic( const scene* ss ) : object( ss )
{
}
#line 62 "nws/quadratic.nw"
quadratic::~quadratic( void ) {
}
#line 75 "nws/quadratic.nw"
void
quadratic::setDimensions( unsigned int dd )
{
    this->squares.resize( dd );
    for ( unsigned int ii=0; ii < dd; ++ii ) {
	this->squares[ ii ].resize( dd, 0.0 );
    }
    this->linears.resize( dd );

    this->object::setDimensions( dd );
}
#line 102 "nws/quadratic.nw"
bool
quadratic::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    
#line 135 "nws/quadratic.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

double t1 = -1.0;
double t2 = -1.0;

double AA = 0.0;
double BB = 0.0;
double CC = this->scalar;

unsigned int len = this->squares.size();

for ( unsigned int ii=0; ii < len; ++ii ) {
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ss = this->squares[ ii ][ jj ];
	AA += ss * rd[ ii ] * rd[ jj ];
	BB += ss * rd[ ii ] * rp[ jj ];
	BB += ss * rp[ ii ] * rd[ jj ];
	CC += ss * rp[ ii ] * rp[ jj ];
    }

    double ll = this->linears[ ii ];
    BB += ll * rd[ ii ];
    CC += ll * rp[ ii ];
}

bool hitQuadratic = false;

if ( ::fabs( AA ) > 0.000001 ) {
    double DD = BB * BB - 4.0 * AA * CC;
    if ( DD > 0.000001 ) {
	double dd = ::sqrt( DD );
	t1 = ( - BB - dd ) / ( 2.0 * AA );
	t2 = ( - BB + dd ) / ( 2.0 * AA );
	if ( t2 < t1 ) {
	    double tmp = t1;
	    t1 = t2;
	    t2 = tmp;
	}
	hitQuadratic = t2 >= 0.000001;
    }
} else if ( ::fabs( BB ) > 0.000001 ) {
    t1 = -CC / BB;
    hitQuadratic = t1 >= 0.000001;
}

#line 110 "nws/quadratic.nw"
    if ( ! hitQuadratic ) {
	if ( this->_inside( rp ) ) {
	    this->forgeHit( direction, start, hit );
	    return true;
	}
	return false;
    }

    if ( t1 >= 0.000001 ) {
	mathvec< double > nn;
	double tt = t1;
	
#line 189 "nws/quadratic.nw"
mathvec< double > spot = rp + ( rd * tt );

nn.resize( len );
for ( unsigned int ii=0; ii < len; ++ii ) {
    double vv = this->linears[ ii ];
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ff = ( this->squares[ ii ][ jj ] + this->squares[ jj ][ ii ] );
	vv += ff * spot[ jj ];
    }
    nn[ ii ] = vv;
}

if ( this->_inside( rp ) ) {
    nn *= -1.0;
    hit.setInside();
} else {
    hit.setOutside();
}

nn.normalize();

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );

#line 122 "nws/quadratic.nw"
    } else if ( t2 >= 0.000001 ) {
	mathvec< double > nn;
	double tt = t2;
	
#line 189 "nws/quadratic.nw"
mathvec< double > spot = rp + ( rd * tt );

nn.resize( len );
for ( unsigned int ii=0; ii < len; ++ii ) {
    double vv = this->linears[ ii ];
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ff = ( this->squares[ ii ][ jj ] + this->squares[ jj ][ ii ] );
	vv += ff * spot[ jj ];
    }
    nn[ ii ] = vv;
}

if ( this->_inside( rp ) ) {
    nn *= -1.0;
    hit.setInside();
} else {
    hit.setOutside();
}

nn.normalize();

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );

#line 126 "nws/quadratic.nw"
    } else {
	return false;
    }

    return true;
}
#line 238 "nws/quadratic.nw"
bool
quadratic::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    
#line 135 "nws/quadratic.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;

double t1 = -1.0;
double t2 = -1.0;

double AA = 0.0;
double BB = 0.0;
double CC = this->scalar;

unsigned int len = this->squares.size();

for ( unsigned int ii=0; ii < len; ++ii ) {
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ss = this->squares[ ii ][ jj ];
	AA += ss * rd[ ii ] * rd[ jj ];
	BB += ss * rd[ ii ] * rp[ jj ];
	BB += ss * rp[ ii ] * rd[ jj ];
	CC += ss * rp[ ii ] * rp[ jj ];
    }

    double ll = this->linears[ ii ];
    BB += ll * rd[ ii ];
    CC += ll * rp[ ii ];
}

bool hitQuadratic = false;

if ( ::fabs( AA ) > 0.000001 ) {
    double DD = BB * BB - 4.0 * AA * CC;
    if ( DD > 0.000001 ) {
	double dd = ::sqrt( DD );
	t1 = ( - BB - dd ) / ( 2.0 * AA );
	t2 = ( - BB + dd ) / ( 2.0 * AA );
	if ( t2 < t1 ) {
	    double tmp = t1;
	    t1 = t2;
	    t2 = tmp;
	}
	hitQuadratic = t2 >= 0.000001;
    }
} else if ( ::fabs( BB ) > 0.000001 ) {
    t1 = -CC / BB;
    hitQuadratic = t1 >= 0.000001;
}

#line 246 "nws/quadratic.nw"
    hits.clear();

    if ( !hitQuadratic ) {
	if ( this->_inside( rp ) ) {
	    intersection* hh = new intersection();
	    hits.insert( hits.end(), hh );
	    this->forgeHit( direction, start, *hh );
	}
    } else {
	if ( t1 >= 0.000001 ) {
	    intersection* h1 = new intersection();
	    hits.insert( hits.end(), h1 );

	    intersection& hit( *h1 );
	    mathvec< double > nn;
	    double tt = t1;
	    
#line 189 "nws/quadratic.nw"
mathvec< double > spot = rp + ( rd * tt );

nn.resize( len );
for ( unsigned int ii=0; ii < len; ++ii ) {
    double vv = this->linears[ ii ];
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ff = ( this->squares[ ii ][ jj ] + this->squares[ jj ][ ii ] );
	vv += ff * spot[ jj ];
    }
    nn[ ii ] = vv;
}

if ( this->_inside( rp ) ) {
    nn *= -1.0;
    hit.setInside();
} else {
    hit.setOutside();
}

nn.normalize();

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );

#line 263 "nws/quadratic.nw"
	    hit.setOutside();
	} 

	if ( t2 >= 0.000001 ) {
	    intersection* h2 = new intersection();
	    hits.insert( hits.end(), h2 );

	    intersection& hit( *h2 );
	    mathvec< double > nn;
	    double tt = t2;
	    
#line 189 "nws/quadratic.nw"
mathvec< double > spot = rp + ( rd * tt );

nn.resize( len );
for ( unsigned int ii=0; ii < len; ++ii ) {
    double vv = this->linears[ ii ];
    for ( unsigned int jj=0; jj < len; ++jj ) {
	double ff = ( this->squares[ ii ][ jj ] + this->squares[ jj ][ ii ] );
	vv += ff * spot[ jj ];
    }
    nn[ ii ] = vv;
}

if ( this->_inside( rp ) ) {
    nn *= -1.0;
    hit.setInside();
} else {
    hit.setOutside();
}

nn.normalize();

hit.setObjectPosition( spot );
hit.setObjectNormal( nn );
hit.setObjectDirection( rd );

hit.setPosition( spot );
hit.setNormal( nn );
hit.setDirection( rd );

hit.reorient( this->scale, this->orientation, this->center );
hit.setDistance( ( hit.getPosition() - start ).magnitude() );

hit.setObjectColor( &this->chroma );

#line 274 "nws/quadratic.nw"
	    hit.setInside();
	}
    }

    return hits.size() > 0;
}
#line 290 "nws/quadratic.nw"
bool
quadratic::inside( const mathvec< double >& start ) const
{
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
quadratic::_inside( const mathvec< double >& rp ) const
{
    unsigned int len = this->center.size();
    double sum = this->scalar;

    for ( unsigned int ii=0; ii < len; ++ii ) {
	double vv = this->linears[ ii ];
	for ( unsigned int jj=0; jj < len; ++jj ) {
	    vv += this->squares[ ii ][ jj ] * rp[ jj ];
	}
	sum += vv * rp[ ii ];
    }

    return sum <= 0.0;
}
#line 326 "nws/quadratic.nw"
std::istream&
operator >> ( std::istream& stream, quadratic& qq )
{
    input in( stream );

    qq.scalar = 0.0;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "squares" ) {
	    stream >> qq.squares >> semicolon;
	} else if ( tok == "linears" ) {
	    stream >> qq.linears >> semicolon;
	} else if ( tok == "scalar" ) {
	    stream >> qq.scalar >> semicolon;
	} else if ( tok == "base" ) {
	    stream >> (object&)qq >> semicolon;
	} else {
	    in.ignore( "QUADRATIC", tok.c_str() );
	}
    }

    qq.linears.resize( qq.squares.size(), 0.0 );
    for ( unsigned int ii=0; ii < qq.squares.size(); ++ii ) {
	qq.squares[ ii ].resize( qq.squares.size(), 0.0 );
    }

    return stream;
}
