#line 343 "nws/csg-set.nw"
#include <string>
#include <sstream>
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "set.h"

const char* csgset::id = "NKVERSION: csgset( 2.1.2007.03.21 )"; 

#line 44 "nws/csg-set.nw"
csgset::csgset( const scene* ss ) : object( ss )
{
}
#line 57 "nws/csg-set.nw"
csgset::~csgset( void ) {
    std::vector< object* >::iterator it;
    for ( it=this->objs.begin(); it != this->objs.end(); ++it ) {
	delete *it;
    }
};
#line 74 "nws/csg-set.nw"
void
csgset::setDimensions( unsigned int dd )
{
    std::vector< object* >::iterator it;
    for ( it=this->objs.begin(); it != this->objs.end(); ++it ) {
	(*it)->setDimensions( dd );
    }
    this->object::setDimensions( dd );
}
#line 99 "nws/csg-set.nw"
bool
csgset::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    intersection& hit
) const {
    std::vector< intersection* > hits;
    bool ret = this->intersect( direction, start, hits );
    if ( ret ) {
	hit = *hits[ 0 ];
	std::vector< intersection* >::iterator it;
	for ( it = hits.begin(); it != hits.end(); ++it ) {
	    delete *it;
	}
    }
    return ret;
}
#line 131 "nws/csg-set.nw"
bool
csgset::intersect(
    const mathvec< double >& direction,
    const mathvec< double >& start,
    std::vector< intersection* >& hits
) const {
    
#line 147 "nws/csg-set.nw"
mathvec< double > rd = direction;
rd.antirotate( this->orientation );
rd /= this->scale;
rd.normalize();

mathvec< double > rp = start - this->center;
rp.antirotate( this->orientation );
rp /= this->scale;
#line 138 "nws/csg-set.nw"
    
#line 158 "nws/csg-set.nw"
unsigned int len = this->objs.size();

std::vector< std::vector< intersection* > > hitLists( len );
std::vector< unsigned int > hitStarts( len );

bool gotHit = false;

for ( unsigned int ii=0; ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    object* obj = this->objs[ ii ];

    if ( obj->intersect( rd, rp, curHits ) ) {
	gotHit = true;
    }
    hitStarts[ ii ] = 0;
}
#line 139 "nws/csg-set.nw"
    if ( gotHit ) {
	
#line 177 "nws/csg-set.nw"
unsigned int bestII = 0;
unsigned int bestJJ = 0;
intersection* bestHit;
do {
    bestHit = 0;
    
#line 190 "nws/csg-set.nw"
bool notSet = true;
double dist = MAXFLOAT;
for ( unsigned int ii=0; ii < len; ++ii ) {
    std::vector< intersection* >& curHits( hitLists[ ii ] );
    unsigned int hStart = hitStarts[ ii ];
    unsigned int hLen = curHits.size();
    if ( hStart < hLen ) {
	intersection* hit = curHits[ hStart ];
	double curDist = hit->getDistance();
	if ( notSet || curDist < dist ) {
	    bestII = ii;
	    bestJJ = hStart;
	    bestHit = hit;
	    dist = curDist;
	    notSet = false;
	}
    }
}
#line 183 "nws/csg-set.nw"
    if ( bestHit != 0 ) {
	
#line 211 "nws/csg-set.nw"
std::vector< intersection* >& curHits( hitLists[ bestII ] );
curHits[ bestJJ ] = 0;
hitStarts[ bestII ] = bestJJ+1;
bestHit->reorient( this->scale, this->orientation, this->center );
bestHit->setDistance( ( bestHit->getPosition() - start ).magnitude() );
if ( this->gotColor ) {
    bestHit->recolor( this->chroma );
}
hits.insert( hits.end(), bestHit );
#line 185 "nws/csg-set.nw"
    }
} while ( bestHit != 0 );
#line 141 "nws/csg-set.nw"
    }
    return gotHit;
}
#line 230 "nws/csg-set.nw"
bool
csgset::inside( const mathvec< double >& start ) const
{
    mathvec< double > rp = start - this->center;
    rp.antirotate( this->orientation );
    rp /= this->scale;

    return this->_inside( rp );
}

bool
csgset::_inside( const mathvec< double >& rp ) const
{
    unsigned int len = this->objs.size();
    bool outside = true;

    for ( unsigned int ii=0; outside && ii < len; ++ii ) {
	outside = ! this->objs[ ii ]->inside( rp );
    }

    return ! outside;
}
#line 262 "nws/csg-set.nw"
std::istream&
operator >> ( std::istream& stream, csgset& cc )
{
    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "object" ) {
	    object* nn = object::read( stream, cc.myScene );
	    if ( nn != 0 ) {
		cc.objs.insert( cc.objs.end(), nn );
	    } else {
		std::cerr << "SET failed to get an object!" << std::endl;
	    }
	} else if ( tok == "base" ) {
	    stream >> (object&)cc >> semicolon;
	} else {
	    in.ignore( "SET", tok.c_str() );
	}
    }

    return stream;
}
