#line 567 "nws/universe.nw"
#include <math.h>
#ifndef MAXFLOAT
#include <values.h>
#endif
#include "input.h"
#include "nkthread.h"
#include "universe.h"

const char* universe::id = "NKVERSION: universe( 2.1.2007.03.21 )"; 

#line 112 "nws/universe-inter.nw"
class intersectionWork : public thread::work {
protected:
    
#line 15 "nws/universe-inter.nw"
const mathvec< double >* position;
const mathvec< double >* direction;
const object* obj;
intersection hit;
bool gotHit;
#line 115 "nws/universe-inter.nw"
public:
    
#line 26 "nws/universe-inter.nw"
intersectionWork( void ) {
};
#line 34 "nws/universe-inter.nw"
virtual ~intersectionWork( void ) {
    this->wait();
};
#line 45 "nws/universe-inter.nw"
void set(
    const mathvec< double >& pp, const mathvec< double >& dd,
    const object* oo
) {
    this->position = &pp;
    this->direction = &dd;
    this->obj = oo;
    this->gotHit = false;
    this->start();
};
#line 63 "nws/universe-inter.nw"
virtual void perform( void ) {
    this->gotHit = this->obj->intersect(
	*this->direction, *this->position, this->hit
    );
    if ( this->gotHit && this->hit.getDistance() > MAXFLOAT / 2.0 ) {
	this->gotHit = false;
    }
};
#line 79 "nws/universe-inter.nw"
bool wasHit( void ) {
    this->wait();
    return this->gotHit;
};
#line 89 "nws/universe-inter.nw"
double getDistance( void ) {
    this->wait();
    return this->hit.getDistance();
};
#line 99 "nws/universe-inter.nw"
const intersection& getIntersection( void ) {
    this->wait();
    return this->hit;
};
#line 117 "nws/universe-inter.nw"
public:
    static const char* id;
};
const char* intersectionWork::id = "NKVERSION: intersectionWork( 2.1.2007.03.21 )"; 
#line 110 "nws/universe-refl.nw"
class reflectionWork : public thread::work {
protected:
    
#line 17 "nws/universe-refl.nw"
const mathvec< double >& reflectedDirection;
const intersection& hit;
unsigned int height;
double contribution;
const universe* univ;
double amount;
mathvec< double > pixel;
#line 113 "nws/universe-refl.nw"
public:
    
#line 36 "nws/universe-refl.nw"
reflectionWork(
    const mathvec< double >& rr,
    const intersection& hh, double cn, unsigned int hg, const universe* uu
) : reflectedDirection( rr ), hit( hh ),
    height( hg ), contribution( cn ), univ( uu )
{
    const color* cc = hit.getColor();
    this->amount = cc->getReflectiveness();
    if ( this->contribution * this->amount > 0.000001 ) {
	this->start();
    } else {
	this->amount = 0.0;
	this->finish();
    }
};
#line 58 "nws/universe-refl.nw"
virtual ~reflectionWork( void ) {
    this->wait();
};
#line 68 "nws/universe-refl.nw"
virtual void perform( void ) {
    mathvec< double > pos = this->hit.getPosition();

    pos += this->reflectedDirection * 0.000001;

    if ( this->univ->evaluate(
	pos, this->reflectedDirection,
	this->contribution, this->height-1, this->pixel
    ) == false ) {
	this->amount = 0.0;
    }
};
#line 88 "nws/universe-refl.nw"
inline double getAmount( void ) {
    this->wait();
    return this->amount;
};
#line 98 "nws/universe-refl.nw"
const mathvec< double >& getPixel( void ) {
    this->wait();
    return this->pixel;
};
#line 115 "nws/universe-refl.nw"
public:
    static const char* id;
};

const char* reflectionWork::id = "NKVERSION: reflectionWork( 2.1.2007.03.21 )"; 
#line 204 "nws/universe-refr.nw"
class refractionWork : public thread::work {
protected:
    
#line 17 "nws/universe-refr.nw"
const intersection& hit;
unsigned int height;
double contribution;
double indexOfRefraction;
const universe* univ;
double amount;
mathvec< double > pixel;
#line 207 "nws/universe-refr.nw"
public:
    
#line 36 "nws/universe-refr.nw"
refractionWork(
    const intersection& hh, double cn, unsigned int hg, double ior,
    const universe* uu
) : hit( hh ), height( hg ), contribution( cn ),
    indexOfRefraction( ior ), univ( uu )
{
    const color* cc = hit.getColor();
    this->amount = cc->getTransparency();
    if ( this->contribution * this->amount > 0.000001 ) {
	this->start();
    } else {
	this->amount = 0.0;
	this->finish();
    }
};
#line 57 "nws/universe-refr.nw"
virtual ~refractionWork( void ) {
    this->wait();
};
#line 129 "nws/universe-refr.nw"
virtual void perform( void ) {
    mathvec< double > pos = this->hit.getPosition();
    mathvec< double > dir = this->hit.getDirection();

    const color* cc = this->hit.getColor();
    double hitIOR = cc->getIndexOfRefraction();

    double ior1;
    double ior2;

    if ( this->hit.isOutside() ) {
	ior1 = this->indexOfRefraction;
	ior2 = hitIOR;
    } else {
	ior1 = hitIOR;
	ior2 = 1.0;
    }

    if ( ::fabs( ior1 - ior2 ) > 0.000001 ) {
	mathvec< double > norm = this->hit.getNormal();
	norm *= -1.0;

	double rn = ( dir ^ norm );

	if ( rn >= -0.000001 ) {
	    double alpha = ior1 / ior2;
	    double beta = - alpha * rn + ::sqrt(
		1.0 + alpha * alpha * ( rn * rn - 1.0 )
	    );

	    dir *= alpha;
	    norm *= beta;
	    dir += norm;
	} else {
	    this->amount = 0.0;
	}
    }

    pos += dir * 0.000001;
    if ( this->univ->evaluate(
	pos, dir, ior2, this->contribution, this->height-1, this->pixel
    ) == false ) {
	this->amount = 0.0;
    }
};
#line 182 "nws/universe-refr.nw"
inline double getAmount( void ) {
    this->wait();
    return this->amount;
};
#line 192 "nws/universe-refr.nw"
const mathvec< double >& getPixel( void ) {
    this->wait();
    return this->pixel;
};
#line 209 "nws/universe-refr.nw"
public:
    static const char* id;
};

const char* refractionWork::id = "NKVERSION: refractionWork( 2.1.2007.03.21 )"; 
#line 204 "nws/universe-shad.nw"
class shadowWork : public thread::work {
protected:
    
#line 20 "nws/universe-shad.nw"
const mathvec< double >* reflectedDirection;
const intersection* hit;
const light* lamp;
double contribution;
unsigned int height;
const universe* univ;
mathvec< double > diffusePixel;
mathvec< double > specularPixel;
double amount;
mathvec< double > pixel;
#line 207 "nws/universe-shad.nw"
public:
    
#line 42 "nws/universe-shad.nw"
shadowWork( void )
{
}
#line 52 "nws/universe-shad.nw"
virtual ~shadowWork( void ) {
    this->wait();
};
#line 62 "nws/universe-shad.nw"
void set(
    const mathvec< double >* rr,
    const intersection* hh, const light* ll,
    double cn, unsigned int hg, const universe* uu
)
{
    this->reflectedDirection = rr;
    this->hit = hh;
    this->lamp = ll;
    this->contribution = cn;
    this->height = hg;
    this->univ = uu;
    this->start();
};
#line 88 "nws/universe-shad.nw"
virtual void perform( void ) {
    
#line 117 "nws/universe-shad.nw"
mathvec< double > pos = this->hit->getPosition();
mathvec< double > dir = this->lamp->getPosition();
dir -= pos;

double lampDist = dir.magnitude();
dir /= lampDist;

pos += dir * 0.000001;

#line 91 "nws/universe-shad.nw"
    mathvec< double > norm = this->hit->getNormal();
    this->amount = norm ^ dir;

    double distFact = ( lampDist * this->lamp->getFallOff() );
    if ( ::fabs( distFact ) > 0.0000001 ) {
	this->amount /= distFact * distFact;
    }

    const color* cc = this->hit->getColor();

    if ( this->amount > 0.0 || cc->getTransparency() > 0.000001 ) {
	this->diffusePixel.resize( cc->getChannelCount(), 0.0 );
	
#line 131 "nws/universe-shad.nw"
this->univ->checkShadow(
    pos, dir, lampDist,
    this->contribution, this->height,
    this->amount, this->diffusePixel
);
#line 104 "nws/universe-shad.nw"
    }

    if ( this->amount > 0.0 ) {
	
#line 141 "nws/universe-shad.nw"
this->lamp->evaluate(
    pos, norm, dir, this->contribution, this->height, this->diffusePixel
);
#line 108 "nws/universe-shad.nw"
	
#line 150 "nws/universe-shad.nw"
this->specularPixel = this->diffusePixel;

double specularness = cc->getSpecularness();
if ( specularness > 0.000001 ) {
    double phong = cc->getPhong();
    double cosAngle = ( *this->reflectedDirection ^ dir );
    double factor = ::pow(cosAngle,phong) * specularness;
    this->specularPixel *= factor;
} else {
    this->specularPixel *= 0.0;
}
#line 109 "nws/universe-shad.nw"
	
#line 167 "nws/universe-shad.nw"
mathvec< double > objPixel;
this->hit->evaluate( this->contribution, this->height, objPixel );
this->diffusePixel *= objPixel;
#line 110 "nws/universe-shad.nw"
    }
};
#line 178 "nws/universe-shad.nw"
inline double getAmount( void ) {
    this->wait();
    return this->amount;
};
#line 189 "nws/universe-shad.nw"
const mathvec< double >& getPixel( double diffuseAmount ) {
    this->wait();
    this->pixel = this->diffusePixel;
    this->pixel *= diffuseAmount;
    this->pixel += this->specularPixel;
    return this->pixel;
};
#line 209 "nws/universe-shad.nw"
public:
    static const char* id;
};

const char* shadowWork::id = "NKVERSION: shadowWork( 2.1.2007.03.21 )"; 
#line 75 "nws/universe.nw"
universe::universe( const scene* ss )
    : myScene( ss ), ambientLight( ss ),
	gotAmbient( false ), skyColor( ss ), quickRender( false )
{
}
#line 90 "nws/universe.nw"
universe::~universe( void )
{
    for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
	delete objects[ ii ];
    }
    for ( unsigned int ii=0; ii < this->lights.size(); ++ii ) {
	delete lights[ ii ];
    }
}
#line 122 "nws/universe.nw"
void
universe::setDimensions( unsigned int dims )
{
    this->dimensions = dims;

    for ( unsigned int ii=0; ii < this->lights.size(); ++ii ) {
	this->lights[ ii ]->setDimensions( this->dimensions );
    }

    for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
	this->objects[ ii ]->setDimensions( this->dimensions );
    }

    this->ambientLight.setDimensions( dims );
    this->skyColor.setDimensions( dims );
}
#line 156 "nws/universe.nw"
bool
universe::evaluate(
    const mathvec< double >& position,
    const mathvec< double >& direction,
    double contribution, unsigned int height,
    mathvec< double >& pixel
) const {
    return this->evaluate(
	position, direction, 1.0, contribution, height, pixel
    );
}
#line 192 "nws/universe.nw"
bool
universe::evaluate(
    const mathvec< double >& position,
    const mathvec< double >& direction,
    double indexOfRefraction,
    double contribution, unsigned int height,
    mathvec< double >& pixel
) const {
    if ( height == 0 || contribution < 0.000001 ) {
	return false;
    } else {
	
#line 223 "nws/universe.nw"
pixel *= 0.0;

#line 265 "nws/universe.nw"
std::vector< intersectionWork > intersections( this->objects.size() );
for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
    intersections[ ii ].set( position, direction, this->objects[ ii ] );
}

#line 227 "nws/universe.nw"
intersectionWork* closest = 0;
double closestDistance = MAXFLOAT;

for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
    if ( intersections[ ii ].wasHit()
    && intersections[ ii ].getDistance() < closestDistance ) {
	closest = &intersections[ ii ];
	closestDistance = closest->getDistance();
    }
}
#line 242 "nws/universe.nw"
if ( closest == 0 ) {
    mathvec< double > normal = direction * -1.0;
    this->skyColor.evaluate(
	position, normal, direction, contribution, height, pixel
    );
    pixel *= contribution;
    return true;
}
#line 255 "nws/universe.nw"
intersection hit = closest->getIntersection();
intersections.resize(0);
#line 204 "nws/universe.nw"
	if ( this->quickRender ) {
	    
#line 276 "nws/universe.nw"
mathvec< double > ndir = hit.getNormal();
double dot = -( ndir ^ direction );
dot = 1.0 - 0.75 * ( 1.0 - dot );
if ( dot < 0.25 ) {
    dot = 0.25;
}
hit.evaluate( contribution, height, pixel );
pixel *= dot;
#line 206 "nws/universe.nw"
	} else {
	    
#line 294 "nws/universe.nw"
mathvec< double > ndir = hit.getNormal();
mathvec< double > rdir = hit.getDirection(); // reflected ray

ndir *= -2.0 * ( rdir ^ ndir );
rdir += ndir;

reflectionWork reflected( rdir, hit, contribution, height, this );
refractionWork refracted( hit, contribution, height, indexOfRefraction, this );
#line 314 "nws/universe.nw"
std::vector< shadowWork > shadows( this->lights.size() );
for ( unsigned int ii=0; ii < shadows.size(); ++ii ) {
    shadows[ ii ].set(
	&rdir, &hit, this->lights[ ii ], contribution, height, this
    );
}
#line 325 "nws/universe.nw"
if ( pixel.size() < hit.getColor()->getChannelCount() ) {
    pixel.resize( hit.getColor()->getChannelCount() );
}

if ( this->gotAmbient ) {
    mathvec< double > normal = direction * -1.0;
    hit.evaluate( contribution, height, pixel );
    mathvec< double > ambientPixel;
    this->ambientLight.evaluate(
	position, normal, direction, contribution, height, ambientPixel
    );
    pixel *= ambientPixel;
} else {
    pixel *= 0.0;
}
#line 348 "nws/universe.nw"
double reflectedAmount = reflected.getAmount();
if ( reflectedAmount > 0.000001 ) {
    pixel += reflectedAmount * reflected.getPixel();
}

double refractedAmount = refracted.getAmount();
if ( refractedAmount > 0.000001 ) {
    pixel += refractedAmount * refracted.getPixel();
}

for ( unsigned int ii=0; ii < shadows.size(); ++ii ) {
    double amount = shadows[ ii ].getAmount();
    if ( amount > 0.000001 ) {
	double diffuseness = 1.0 - reflectedAmount - refractedAmount;
	pixel += amount * shadows[ ii ].getPixel( diffuseness );
    }
}

pixel *= contribution;
#line 208 "nws/universe.nw"
	}
	return true;
    }
}
#line 395 "nws/universe.nw"
void
universe::checkShadow(
    const mathvec< double >& position,
    const mathvec< double >& direction,
    double lampDistance,
    double contribution,
    unsigned int height,
    double& amount,
    mathvec< double >& pixel
) const
{
    
#line 265 "nws/universe.nw"
std::vector< intersectionWork > intersections( this->objects.size() );
for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
    intersections[ ii ].set( position, direction, this->objects[ ii ] );
}
#line 407 "nws/universe.nw"
    
#line 416 "nws/universe.nw"
for ( unsigned int ii=0; ii < this->objects.size(); ++ii ) {
    if ( intersections[ ii ].wasHit()
    && intersections[ ii ].getDistance() < lampDistance ) {
	const color* cc = intersections[ ii ].getIntersection().getColor();
	double tt = cc->getTransparency();
	if ( tt < 0.000001 ) {
	    amount = 0.0;
	    break;
	} else {
	    mathvec< double > objPixel;
	    if ( tt < 0.999999 ) {
		intersections[ ii ].getIntersection().evaluate(
		    contribution * tt, height, objPixel
		);
		objPixel *= ( 1.0 - tt );
		objPixel *= pixel;
		pixel *= tt;
		pixel += objPixel;
	    } else {
		pixel *= tt;
	    }
	    amount *= tt;
	}
    }
}
#line 408 "nws/universe.nw"
}
#line 464 "nws/universe.nw"
std::istream&
operator >> ( std::istream& stream, universe& uu )
{
    input in( stream );
    unsigned int dims = 2;

    uu.gotAmbient = false;

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "dimensions" || tok == "dims" ) {
	    stream >> dims >> semicolon;
	} else if ( tok == "ambientLight" || tok == "ambient" ) {
	    stream >> uu.ambientLight >> semicolon;
	    uu.gotAmbient = true;
	} else if ( tok == "skyColor" || tok == "sky" ) {
	    stream >> uu.skyColor >> semicolon;
	} else if ( tok == "quickRender" ) {
	    stream >> semicolon;
	    uu.quickRender = true;
	} else if ( tok == "light" ) {
	    light* nn = new light( uu.myScene );
	    stream >> *nn >> semicolon;
	    uu.lights.insert( uu.lights.end(), nn );
	} else if ( tok == "object" ) {
	    object* nn = object::read( stream, uu.myScene );
	    if ( nn != 0 ) {
		uu.objects.insert( uu.objects.end(), nn );
	    }
	} else {
	    in.ignore( "UNIVERSE", tok.c_str() );
	}
    }

    uu.setDimensions( dims );

    return stream;
}
