#line 791 "nws/view.nw"
#include <math.h>
#include "input.h"
#include "view.h"

const char* view::id = "NKVERSION: view( 2.1.2007.03.21 )"; 

#line 230 "nws/view.nw"
view::view( const scene* ss )
    : myScene( ss ), viewport( ss ), 
#line 221 "nws/view.nw"
wrap( false ), gamma( 1.0 ), border( 1 ), composite( 0 ), traceDepth( 12 )
#line 232 "nws/view.nw"
{
}
#line 245 "nws/view.nw"
view::~view( void )
{
    for ( unsigned int ii=0; ii < this->images.size(); ++ii ) {
	delete this->images[ ii ];
    }
}
#line 268 "nws/view.nw"
void
view::render( void ) const
{
    
#line 302 "nws/view.nw"
mathvec< unsigned int > coordinates( 0U, this->size.size() );
#line 272 "nws/view.nw"
    
#line 312 "nws/view.nw"
unsigned int width = 1;
unsigned int height = 1;
unsigned int wdirs = 0;
unsigned int hdirs = 0;
for ( unsigned int ii=0; ii < this->composite; ++ii ) {
    bool bb = this->contributesToWidth[ ii ];
    if ( bb ) {
	width  =  width * this->size[ ii ];
	++wdirs;
    } else {
	height = height * this->size[ ii ];
	++hdirs;
    }
}
if ( wdirs < 1 ) {
    width += 2 * this->border;
}
if ( hdirs < 1 ) {
    height += 2 * this->border;
}
#line 273 "nws/view.nw"
    
#line 337 "nws/view.nw"
unsigned int depth=1;
for ( unsigned int ii=this->composite; ii < this->size.size(); ++ii ) {
    depth *= this->size[ ii ];
}
#line 274 "nws/view.nw"
    
#line 347 "nws/view.nw"
for ( unsigned int ii=0; ii < this->images.size(); ++ii ) {
    this->images[ ii ]->setSize(
	width, height, depth,
	this->viewport.getChannelCount()
    );
}

#line 276 "nws/view.nw"
    mathvec< double > position( 0.0, this->size.size() + 1 );
    mathvec< double > direction( this->size.size() + 1 );
    mathvec< double > normal;
    mathvec< double > pixel;

    for ( unsigned int kk=0; kk < depth; ++kk ) {
	for ( unsigned int jj=0; jj < height; ++jj ) {
	    for ( unsigned int ii=0; ii < width; ++ii ) {
		if ( this->calculateDirection( coordinates, direction ) ) {
		    
#line 358 "nws/view.nw"
normal = direction * -1.0;
#line 286 "nws/view.nw"
		    
#line 364 "nws/view.nw"
this->viewport.evaluate(
    position, normal, direction,
    1.0, this->traceDepth,
    pixel
);

unsigned int len = pixel.size();
for ( unsigned int ii=0; ii < len; ++ii ) {
    double val = pixel[ ii ];
    if ( val > 0.0 ) {
	pixel[ ii ] = ::pow( val, this->gamma );
    }
}
#line 287 "nws/view.nw"
		    
#line 382 "nws/view.nw"
for ( unsigned int ll=0; ll < this->images.size(); ++ll ) {
    this->images[ ll ]->add( pixel );
}
#line 288 "nws/view.nw"
		} else {
		    
#line 390 "nws/view.nw"
for ( unsigned int ll=0; ll < this->images.size(); ++ll ) {
    this->images[ ll ]->addBorderPixel();
}
#line 290 "nws/view.nw"
		}
		this->incrementDirection( coordinates );
	    }
	}
    }
}
#line 418 "nws/view.nw"
bool
view::calculateDirection(
    const mathvec< unsigned int >& coordinates,
    mathvec< double >& direction
) const {
    unsigned int dims = coordinates.size();

    direction[ 0 ] = 0.0;

    
#line 438 "nws/view.nw"
for ( unsigned int ii=0; ii < dims; ++ii ) {
    unsigned int index = this->permutation[ ii ] + 1;

    unsigned int bb = 0;
    if ( this->needsBorders[ ii ] ) {
	bb = this->border;
    }

    double is = (double)( this->size[ ii ] - 2*bb );
    double ss = is / 2.0 - 0.5;

    if ( ss < 0.000001 ) {
	direction[ index ] = 0.0;
    } else {
	unsigned int cc = coordinates[ ii ];
	double tt = (double)( cc - bb ) - ss;

	if ( cc < bb || cc + bb >= this->size[ ii ] ) {
	    return false;
	}

	direction[ index ] = this->angularExtents[ ii ] * tt / ss;
    }
}
#line 428 "nws/view.nw"
    
#line 473 "nws/view.nw"
double mag = direction.magnitude();

if ( this->wrap == false && mag > M_PI ) {
    return false;
} else {
    if ( mag < 0.000001 ) {
	direction *= 0.0;
	direction[ 0 ] = 1.0;
    } else {
	direction *= -::sin( mag ) / mag;
	direction[ 0 ] = ::cos( mag );
    }
    return true;
}
#line 429 "nws/view.nw"
}
#line 506 "nws/view.nw"
void
view::incrementDirection( mathvec< unsigned int >& coordinates ) const
{
    unsigned int dims = coordinates.size();

    for ( unsigned int ii=0; ii < dims; ++ii ) {
	if ( ++coordinates[ ii ] < this->size[ ii ] ) {
	    break;
	} else {
	    coordinates[ ii ] = 0;
	}
    }
}
#line 550 "nws/view.nw"
std::istream&
operator >> ( std::istream& stream, view& vv )
{
    mathvec< double > units;
    mathvec< double > dpu;
    mathvec< double > fishEye;
    double viewDepth = 12.0;

    vv.vertical = false;

    input in( stream );

    while ( in.check() ) {
	char token[ input::MAX_TOKEN_LEN ];

	if ( in.get( token, sizeof(token) ) == 0 ) {
	    return stream;
	}

	std::string tok( token );
	char semicolon;

	if ( tok == "unitsPerDimension" || tok == "units" ) {
	    stream >> units >> semicolon;
	} else if ( tok == "dotsPerUnit" || tok == "dpu" || tok == "dpi" ) {
	    stream >> dpu >> semicolon;
	} else if ( tok == "viewDepth" || tok == "depth" ) {
	    stream >> viewDepth >> semicolon;
	} else if ( tok == "fishEye" ) {
	    stream >> fishEye >> semicolon;
	} else if ( tok == "wrapAround" || tok == "wrap" ) {
	    vv.wrap = true;
	    stream >> semicolon;
	} else if ( tok == "gamma" ) {
	    stream >> vv.gamma >> semicolon;
	} else if ( tok == "border" ) {
	    stream >> vv.border >> semicolon;
	} else if ( tok == "composite" ) {
	    stream >> vv.composite >> semicolon;
	} else if ( tok == "vertical" ) {
	    vv.vertical = true;
	    stream >> semicolon;
	} else if ( tok == "traceDepth" || tok == "maxDepth" ) {
	    stream >> vv.traceDepth >> semicolon;
	} else if ( tok == "viewport" || tok == "color" ) {
	    stream >> vv.viewport >> semicolon;
	} else if ( tok == "image" ) {
	    image* nn = image::read( stream );
	    vv.images.insert( vv.images.end(), nn );
	    stream >> semicolon;
	} else {
	    in.ignore( "VIEW", tok.c_str() );
	}
    }

    
#line 620 "nws/view.nw"
unsigned int dims = units.size();
if ( dims == 0 ) {
    std::cerr << "View: length dimensions are zero" << std::endl;
    return stream;
}

dpu.resize( dims, 1.0 );

if ( vv.composite == 0 || vv.composite > dims ) {
    vv.composite = dims;
}

unsigned int ii;
fishEye.resize( dims, 0.0 );
#line 606 "nws/view.nw"
    
#line 645 "nws/view.nw"
vv.permutation.resize( dims );
vv.contributesToWidth.resize( vv.composite );
vv.needsBorders.resize( dims, false );

unsigned int evenDims = vv.composite / 2;
unsigned int oddDims = vv.composite - evenDims;

for (ii = 0; ii < oddDims; ++ii) {
    vv.permutation[ii] = ii * 2;
    vv.contributesToWidth[ ii ] = true;
}

for (ii = oddDims; ii < vv.composite; ++ii) {
    vv.permutation[ii] = ( ii - oddDims ) * 2 + 1;
    vv.contributesToWidth[ ii ] = false;
}

for (ii = vv.composite; ii < dims; ++ii) {
    vv.permutation[ii] = ii;
}

if ( vv.vertical && oddDims > 0 && evenDims > 0 ) {
    
#line 696 "nws/view.nw"
if ( ( vv.composite & 1 ) == 1 ) {
    unsigned int tt = vv.permutation[ oddDims - 1 ];
    for ( ii=oddDims; ii < vv.composite; ++ii ) {
	vv.permutation[ ii - 1 ] = vv.permutation[ ii ];
    }
    vv.permutation[ vv.composite-1 ] = tt;
    vv.contributesToWidth[ oddDims - 1 ] = false;
    --oddDims;
    ++evenDims;
}
for ( ii=1; ii < oddDims; ++ii ) {
    unsigned int tt = vv.permutation[ ii ];
    vv.permutation[ ii ] = vv.permutation[ ii + oddDims ];
    vv.permutation[ ii + oddDims ] = tt;
}
#line 668 "nws/view.nw"
}

unsigned int wdims = 0;
unsigned int hdims = 0;
for ( ii = 0; ii < vv.composite; ++ii ) {
    if ( vv.contributesToWidth[ ii ] ) {
	++wdims;
    } else {
	++hdims;
    }
}

for ( ii = 0; ii < vv.composite; ++ii ) {
    if ( vv.contributesToWidth[ ii ] ) {
	vv.needsBorders[ ii ] = ( --wdims > 0 );
    } else {
	vv.needsBorders[ ii ] = ( --hdims > 0 );
    }
}
for ( ii = vv.composite; ii < dims; ++ii ) {
    vv.needsBorders[ ii ] = false;
}
#line 607 "nws/view.nw"
    
#line 721 "nws/view.nw"
vv.size.resize( dims );
vv.angularExtents.resize( dims );

for (ii = 0; ii < dims; ++ii) {
    unsigned int permuted = vv.permutation[ii];
    double un = units[permuted];
    double len = dpu[permuted] * un;
    double baseAngle;

    vv.size[ ii ] = (unsigned int)(::fabs( len ) + 0.5);
    if ( permuted+1 != vv.composite && permuted+2 != vv.composite ) {
	vv.size[ ii ] += 2 * vv.border;
    }

    baseAngle = atan2( un, 2.0 * viewDepth );
    vv.angularExtents[ ii ] = baseAngle * ( 1.0 +  fishEye[permuted]);
}

#line 609 "nws/view.nw"
    vv.viewport.setDimensions( dims+1 );

    return stream;
}
